import pathlib

from jschon.catalog import Catalog, LocalSource
from jschon.uri import URI
from jschon.vocabulary.annotation import *
from jschon.vocabulary.applicator import *
from jschon.vocabulary.core import *
from jschon.vocabulary.format import *
from jschon.vocabulary.legacy import *
from jschon.vocabulary.validation import *


def initialize(catalog: Catalog):
    catalog.add_uri_source(
        URI('https://json-schema.org/draft/2019-09/'),
        LocalSource(pathlib.Path(__file__).parent / 'json-schema-spec-2019-09', suffix='.json'),
    )

    catalog.create_vocabulary(
        URI("https://json-schema.org/draft/2019-09/vocab/core"),
        SchemaKeyword,
        VocabularyKeyword,
        IdKeyword,
        RefKeyword,
        AnchorKeyword,
        RecursiveRefKeyword_2019_09,
        RecursiveAnchorKeyword_2019_09,
        DefsKeyword,
        CommentKeyword,
    )

    catalog.create_vocabulary(
        URI("https://json-schema.org/draft/2019-09/vocab/applicator"),
        AllOfKeyword,
        AnyOfKeyword,
        OneOfKeyword,
        NotKeyword,
        IfKeyword,
        ThenKeyword,
        ElseKeyword,
        DependentSchemasKeyword,
        ItemsKeyword_2019_09,
        AdditionalItemsKeyword_2019_09,
        UnevaluatedItemsKeyword_2019_09,
        ContainsKeyword,
        PropertiesKeyword,
        PatternPropertiesKeyword,
        AdditionalPropertiesKeyword,
        UnevaluatedPropertiesKeyword,
        PropertyNamesKeyword,
    )

    catalog.create_vocabulary(
        URI("https://json-schema.org/draft/2019-09/vocab/validation"),
        TypeKeyword,
        EnumKeyword,
        ConstKeyword,
        MultipleOfKeyword,
        MaximumKeyword,
        ExclusiveMaximumKeyword,
        MinimumKeyword,
        ExclusiveMinimumKeyword,
        MaxLengthKeyword,
        MinLengthKeyword,
        PatternKeyword,
        MaxItemsKeyword,
        MinItemsKeyword,
        UniqueItemsKeyword,
        MaxContainsKeyword,
        MinContainsKeyword,
        MaxPropertiesKeyword,
        MinPropertiesKeyword,
        RequiredKeyword,
        DependentRequiredKeyword,
    )

    catalog.create_vocabulary(
        URI("https://json-schema.org/draft/2019-09/vocab/format"),
        FormatKeyword,
    )

    catalog.create_vocabulary(
        URI("https://json-schema.org/draft/2019-09/vocab/meta-data"),
        TitleKeyword,
        DescriptionKeyword,
        DefaultKeyword,
        DeprecatedKeyword,
        ReadOnlyKeyword,
        WriteOnlyKeyword,
        ExamplesKeyword,
    )

    catalog.create_vocabulary(
        URI("https://json-schema.org/draft/2019-09/vocab/content"),
        ContentMediaTypeKeyword,
        ContentEncodingKeyword,
        ContentSchemaKeyword,
    )

    catalog.create_metaschema(
        URI("https://json-schema.org/draft/2019-09/schema"),
        URI("https://json-schema.org/draft/2019-09/vocab/core"),
        URI("https://json-schema.org/draft/2019-09/vocab/applicator"),
        URI("https://json-schema.org/draft/2019-09/vocab/validation"),
        URI("https://json-schema.org/draft/2019-09/vocab/format"),
        URI("https://json-schema.org/draft/2019-09/vocab/meta-data"),
        URI("https://json-schema.org/draft/2019-09/vocab/content"),
    )



from __future__ import annotations

from copy import deepcopy
from enum import Enum
from typing import Dict, Iterable, List, Mapping, MutableSequence, Optional, Sequence, Union, overload

from jschon.exceptions import JSONPatchError, JSONPointerError
from jschon.json import JSON, JSONCompatible
from jschon.jsonpointer import JSONPointer

__all__ = [
    'JSONPatch',
    'JSONPatchOperation',
    'PatchOp',
    'apply_add',
    'apply_remove',
    'apply_replace',
    'apply_move',
    'apply_copy',
    'apply_test',
]


class PatchOp(str, Enum):
    ADD = 'add'
    REMOVE = 'remove'
    REPLACE = 'replace'
    MOVE = 'move'
    COPY = 'copy'
    TEST = 'test'

    def __repr__(self) -> str:
        return f'PatchOp.{self.name}'


class JSONPatchOperation:
    """:rfc:`6902`-conformant JSON patch operation object."""

    def __new__(
            cls,
            *,
            op: PatchOp,
            path: Union[str, JSONPointer],
            value: JSONCompatible = None,
            from_: Optional[Union[str, JSONPointer]] = None,
            **kwargs: Union[str, JSONPointer],
    ) -> JSONPatchOperation:
        """Create and return a new :class:`JSONPatchOperation` instance.

        :param op: The operation to perform. One of ``add``, ``remove``,
            ``replace``, ``move``, ``copy``, ``test``.
        :param path: A JSON pointer to the target location.
        :param value: For ``add`` and ``replace`` operations, the value
            to set at the target location. For ``test``, the value to
            compare with the target.
        :param from_: The location from which to ``move`` or ``copy``.
            An alias for `from`, which may be passed via `kwargs`.
        """
        self = object.__new__(cls)
        self.op = PatchOp(op)
        self.path = JSONPointer(path) if isinstance(path, str) else path
        self.value = value
        if from_ is None:
            from_ = kwargs.pop('from', None)
        self.from_ = JSONPointer(from_) if isinstance(from_, str) else from_
        return self

    def apply(self, document: JSONCompatible) -> JSONCompatible:
        """Apply the patch operation to `document` and return the
        resultant document."""
        if self.op == 'add':
            return apply_add(document, self.path, self.value)
        if self.op == 'remove':
            return apply_remove(document, self.path)
        if self.op == 'replace':
            return apply_replace(document, self.path, self.value)
        if self.op == 'move':
            return apply_move(document, self.path, self.from_)
        if self.op == 'copy':
            return apply_copy(document, self.path, self.from_)
        if self.op == 'test':
            return apply_test(document, self.path, self.value)

    def asdict(self) -> Dict[str, JSONCompatible]:
        """Return `self` as a dict."""
        result = {
            'op': self.op.value,
            'path': str(self.path),
        }
        if self.op in ('add', 'replace', 'test'):
            result['value'] = self.value
        elif self.op in ('move', 'copy'):
            result['from'] = str(self.from_)

        return result

    def __eq__(self, other: Union[JSONPatchOperation, Mapping[str, JSONCompatible]]) -> bool:
        """Return `self == other`."""
        if not isinstance(other, JSONPatchOperation):
            other = JSONPatchOperation(**other)
        return (self.op == other.op and
                self.path == other.path and
                self.from_ == other.from_ and
                JSON(self.value) == JSON(other.value))

    def __repr__(self) -> str:
        """Return `repr(self)`."""
        return f'JSONPatchOperation(op={self.op!r}, path={self.path!r}, from_={self.from_!r}, value={self.value!r})'


class JSONPatch(MutableSequence[JSONPatchOperation]):
    """:rfc:`6902`-conformant JSON Patch implementation."""

    def __init__(self, *operations: Union[JSONPatchOperation, Mapping[str, JSONCompatible]]) -> None:
        """Initialize a :class:`JSONPatch` instance from the given `operations`,
        each of which may be a :class:`JSONPatchOperation` or a JSON patch
        operation-conformant dictionary.
        """
        self._operations: List[JSONPatchOperation] = [
            operation if isinstance(operation, JSONPatchOperation) else JSONPatchOperation(**operation)
            for operation in operations
        ]

    def evaluate(self, document: JSONCompatible) -> JSONCompatible:
        """Return the result of sequentially applying all patch operations
        to `document`, as a new document. `document` itself is not modified."""
        result = deepcopy(document)
        for operation in self._operations:
            result = operation.apply(result)
        return result

    def aslist(self) -> List[Dict[str, JSONCompatible]]:
        """Return `self` as a list of operation dicts."""
        return [
            operation.asdict()
            for operation in self._operations
        ]

    @overload
    def __getitem__(self, index: int) -> JSONPatchOperation:
        ...

    @overload
    def __getitem__(self, index: slice) -> JSONPatch:
        ...

    def __getitem__(self, index):
        """Return `self[index]`."""
        if isinstance(index, int):
            return self._operations[index]
        if isinstance(index, slice):
            return JSONPatch(*self._operations[index])
        raise TypeError('Expecting int or slice')

    def __setitem__(self, index: int, operation: Union[JSONPatchOperation, Mapping[str, JSONCompatible]]) -> None:
        """Set `self[index]` to `operation`."""
        if not isinstance(operation, JSONPatchOperation):
            operation = JSONPatchOperation(**operation)
        self._operations[index] = operation

    def __delitem__(self, index: int) -> None:
        """Delete `self[index]`."""
        del self._operations[index]

    def __len__(self) -> int:
        """Return `len(self)`."""
        return len(self._operations)

    def insert(self, index: int, operation: Union[JSONPatchOperation, Mapping[str, JSONCompatible]]) -> None:
        """Insert `operation` before `index`."""
        if not isinstance(operation, JSONPatchOperation):
            operation = JSONPatchOperation(**operation)
        self._operations.insert(index, operation)

    def __eq__(self, other: Union[JSONPatch, Iterable[Union[JSONPatchOperation, Mapping[str, JSONCompatible]]]]) -> bool:
        """Return `self == other`."""
        if not isinstance(other, JSONPatch):
            other = JSONPatch(*other)
        return self._operations == other._operations

    def __repr__(self) -> str:
        """Return `repr(self)`."""
        return f'JSONPatch(*{self._operations!r})'


class NodeType(Enum):
    ROOT = 0
    ARRAY_ITEM = 1
    ARRAY_ITEM_NEW = 2
    OBJECT_PROPERTY = 3
    OBJECT_PROPERTY_NEW = 4


class Node:
    def __init__(self, document: JSONCompatible, path: JSONPointer):
        if not path:
            self.type = NodeType.ROOT
            self.parent = None
            self.index = None
            return

        try:
            self.parent = (parent := path[:-1].evaluate(document))
            key = path[-1]
        except JSONPointerError as e:
            raise JSONPatchError(f'Expecting an array or object at {path[:-1]}') from e

        if isinstance(parent, Sequence):
            try:
                if key == '-' or int(key) == len(parent):
                    self.type = NodeType.ARRAY_ITEM_NEW
                    self.index = len(parent)
                elif 0 <= int(key) < len(parent):
                    self.type = NodeType.ARRAY_ITEM
                    self.index = int(key)
                else:
                    raise ValueError
            except ValueError:
                raise JSONPatchError(f'Invalid array index {key}')

        elif isinstance(parent, Mapping):
            self.type = NodeType.OBJECT_PROPERTY if key in parent \
                else NodeType.OBJECT_PROPERTY_NEW
            self.index = key

        else:
            assert False


def apply_add(document: JSONCompatible, path: JSONPointer, value: JSONCompatible) -> JSONCompatible:
    target = Node(document, path)
    value = deepcopy(value)
    if target.type == NodeType.ROOT:
        return value

    if target.type in (NodeType.ARRAY_ITEM, NodeType.ARRAY_ITEM_NEW):
        target.parent.insert(target.index, value)

    elif target.type in (NodeType.OBJECT_PROPERTY, NodeType.OBJECT_PROPERTY_NEW):
        target.parent[target.index] = value

    return document


def apply_remove(document: JSONCompatible, path: JSONPointer) -> JSONCompatible:
    target = Node(document, path)
    if target.type == NodeType.ROOT:
        return None

    if target.type in (NodeType.ARRAY_ITEM, NodeType.OBJECT_PROPERTY):
        del target.parent[target.index]

    else:
        raise JSONPatchError(f'Cannot remove nonexistent target at {path}')

    return document


def apply_replace(document: JSONCompatible, path: JSONPointer, value: JSONCompatible) -> JSONCompatible:
    target = Node(document, path)
    value = deepcopy(value)
    if target.type == NodeType.ROOT:
        return value

    if target.type in (NodeType.ARRAY_ITEM, NodeType.OBJECT_PROPERTY):
        target.parent[target.index] = value

    else:
        raise JSONPatchError(f'Cannot replace nonexistent target at {path}')

    return document


def apply_move(document: JSONCompatible, path: JSONPointer, from_: JSONPointer) -> JSONCompatible:
    try:
        value = from_.evaluate(document)
    except JSONPointerError as e:
        raise JSONPatchError(f'Cannot move from nonexistent location {from_}') from e

    document = apply_remove(document, from_)
    return apply_add(document, path, value)


def apply_copy(document: JSONCompatible, path: JSONPointer, from_: JSONPointer) -> JSONCompatible:
    try:
        value = from_.evaluate(document)
    except JSONPointerError as e:
        raise JSONPatchError(f'Cannot copy from nonexistent location {from_}') from e

    return apply_add(document, path, value)


def apply_test(document: JSONCompatible, path: JSONPointer, value: JSONCompatible) -> JSONCompatible:
    target = Node(document, path)
    if target.type in (NodeType.ROOT, NodeType.ARRAY_ITEM, NodeType.OBJECT_PROPERTY):
        if JSON(path.evaluate(document)) != JSON(value):
            raise JSONPatchError(f'The value at {path} does not equal {value}')

    else:
        raise JSONPatchError(f'Cannot test nonexistent target at {path}')

    return document



from __future__ import annotations

from collections import deque
from contextlib import contextmanager
from functools import cached_property
from typing import Any, ContextManager, Dict, Hashable, Iterator, Mapping, Optional, TYPE_CHECKING, Tuple, Type, Union
from uuid import uuid4

from jschon.exceptions import JSONSchemaError
from jschon.json import JSON, JSONCompatible
from jschon.jsonpointer import JSONPointer
from jschon.uri import URI

if TYPE_CHECKING:
    from jschon.catalog import Catalog
    from jschon.vocabulary import Keyword, KeywordClass, Metaschema

__all__ = [
    'JSONSchema',
    'Result',
]


class JSONSchema(JSON):
    """JSON schema document model."""

    def __init__(
            self,
            value: Union[bool, Mapping[str, JSONCompatible]],
            *,
            catalog: Union[str, Catalog] = 'catalog',
            session: Hashable = 'default',
            uri: URI = None,
            metaschema_uri: URI = None,
            parent: JSON = None,
            key: str = None,
    ):
        """Initialize a :class:`JSONSchema` instance from the given
        schema-compatible `value`.

        :param value: a schema-compatible Python object
        :param catalog: the catalog in which the schema is cached
        :param session: a session identifier, identifying which session
            cache to put the schema in
        :param uri: the URI identifying the schema; an ``"$id"`` keyword
            appearing in `value` will override this
        :param metaschema_uri: the URI identifying the schema's metaschema;
            a ``"$schema"`` keyword appearing in `value` will override this
        :param parent: the parent node of the schema; used internally when
            creating a subschema
        :param key: the index of the schema within its parent; used internally
            when creating a subschema
        """
        from jschon.catalog import Catalog

        if not isinstance(catalog, Catalog):
            catalog = Catalog.get_catalog(catalog)

        self.catalog: Catalog = catalog
        """The catalog in which the schema is cached."""

        self.session: Hashable = session
        """A session identifier, identifying which session
        cache to use for the schema."""

        if uri is not None:
            catalog.add_schema(uri, self, session=session)

        self._uri: Optional[URI] = uri
        self._metaschema_uri: Optional[URI] = metaschema_uri

        self.keywords: Dict[str, Keyword] = {}
        """A dictionary of the schema's :class:`~jschon.vocabulary.Keyword`
        objects, indexed by keyword name."""

        # do not call super().__init__
        # all inherited attributes are initialized here:

        self.type: str
        """The JSON type of the schema. One of ``"boolean"``, ``"object"``."""

        self.data: Union[bool, Dict[str, JSON]]
        """The schema data.
        
        =========   ===============
        JSON type   data type
        =========   ===============
        boolean     bool
        object      dict[str, JSON]
        =========   ===============
        """

        self.parent: Optional[JSON] = parent
        """The containing :class:`~jschon.json.JSON` or :class:`JSONSchema` node."""

        self.key: Optional[str] = key
        """The index of the schema within its parent."""

        if isinstance(value, bool):
            self.type = "boolean"
            self.data = value

        elif isinstance(value, Mapping):
            self.type = "object"
            self.data = {}

            if self.parent is None and self.uri is None:
                self.uri = URI(f'urn:uuid:{uuid4()}')

            self._bootstrap(value)

            kwclasses = {
                key: kwclass for key in value
                if (key not in self.keywords and  # skip bootstrapped keywords
                    (kwclass := self.metaschema.kwclasses.get(key)))
            }

            for kwclass in self._resolve_dependencies(kwclasses):
                kw = kwclass(self, value[(key := kwclass.key)])
                self.keywords[key] = kw
                self.data[key] = kw.json

            if self.parent is None:
                self._resolve_references()

        else:
            raise TypeError(f"{value=} is not JSONSchema-compatible")

    def _bootstrap(self, value: Mapping[str, JSONCompatible]) -> None:
        from jschon.vocabulary.core import IdKeyword, SchemaKeyword, VocabularyKeyword
        boostrap_kwclasses = {
            "$id": IdKeyword,
            "$schema": SchemaKeyword,
            "$vocabulary": VocabularyKeyword,
        }
        for key, kwclass in boostrap_kwclasses.items():
            if key in value:
                kw = kwclass(self, value[key])
                self.keywords[key] = kw
                self.data[key] = kw.json

    def _resolve_references(self) -> None:
        for kw in self.keywords.values():
            if hasattr(kw, 'resolve'):
                kw.resolve()
            elif isinstance(kw.json, JSONSchema):
                kw.json._resolve_references()
            elif kw.json.type == "array":
                for item in kw.json:
                    if isinstance(item, JSONSchema):
                        item._resolve_references()
            elif kw.json.type == "object":
                for item in kw.json.values():
                    if isinstance(item, JSONSchema):
                        item._resolve_references()

    @staticmethod
    def _resolve_dependencies(kwclasses: Dict[str, KeywordClass]) -> Iterator[KeywordClass]:
        dependencies = {
            kwclass: [depclass for dep in kwclass.depends_on
                      if (depclass := kwclasses.get(dep))]
            for kwclass in kwclasses.values()
        }
        while dependencies:
            for kwclass, depclasses in dependencies.items():
                if not depclasses:
                    del dependencies[kwclass]
                    for deps in dependencies.values():
                        try:
                            deps.remove(kwclass)
                        except ValueError:
                            pass
                    yield kwclass
                    break

    def validate(self) -> Result:
        """Validate the schema against its metaschema."""
        return self.metaschema.evaluate(self)

    def evaluate(self, instance: JSON, result: Result = None) -> Result:
        """Evaluate a JSON document and return the evaluation result.

        :param instance: the JSON document to evaluate
        :param result: the current result node; given by keywords
            when invoking this method recursively
        """
        if result is None:
            result = Result(self, instance)

        if self.data is True:
            pass

        elif self.data is False:
            result.fail("The instance is disallowed by a boolean false schema")

        else:
            for key, keyword in self.keywords.items():
                if not keyword.static and keyword.can_evaluate(instance):
                    with result(instance, key, self) as subresult:
                        keyword.evaluate(instance, subresult)

            if any(
                    not child.passed
                    for child in result.children.values()
                    if child.instance.path == instance.path
            ):
                result.fail()

        return result

    @cached_property
    def parentschema(self) -> Optional[JSONSchema]:
        """The containing :class:`JSONSchema` instance.
        
        Note that this is not necessarily the same as `self.parent`.
        """
        parent = self.parent
        while parent is not None:
            if isinstance(parent, JSONSchema):
                return parent
            parent = parent.parent

    @property
    def metaschema(self) -> Metaschema:
        """The schema's :class:`~jschon.vocabulary.Metaschema`."""
        from jschon.vocabulary import Metaschema

        if (uri := self.metaschema_uri) is None:
            raise JSONSchemaError("The schema's metaschema URI has not been set")

        if not isinstance(
                metaschema := self.catalog.get_schema(uri, session='__meta__'),
                Metaschema,
        ):
            raise JSONSchemaError(f"The schema referenced by {uri} is not a metachema")

        return metaschema

    @property
    def metaschema_uri(self) -> Optional[URI]:
        """The :class:`~jschon.uri.URI` identifying the schema's metaschema.
        
        If not defined on this (sub)schema, the metaschema URI
        is determined by the parent schema.
        """
        if self._metaschema_uri is not None:
            return self._metaschema_uri
        if self.parentschema is not None:
            return self.parentschema.metaschema_uri

    @metaschema_uri.setter
    def metaschema_uri(self, value: Optional[URI]) -> None:
        self._metaschema_uri = value

    @property
    def base_uri(self) -> Optional[URI]:
        """The schema's base :class:`~jschon.uri.URI`.
        
        The base URI is obtained by searching up the schema tree
        for a schema URI, and removing any fragment.
        """
        if self._uri is not None:
            return self._uri.copy(fragment=False)
        if self.parentschema is not None:
            return self.parentschema.base_uri

    @property
    def uri(self) -> Optional[URI]:
        """The :class:`~jschon.uri.URI` identifying the schema.

        Used as the key for caching the schema in the catalog.
        """
        return self._uri

    @uri.setter
    def uri(self, value: Optional[URI]) -> None:
        if self._uri != value:
            if self._uri is not None:
                self.catalog.del_schema(self._uri, session=self.session)

            self._uri = value

            if self._uri is not None:
                self.catalog.add_schema(self._uri, self, session=self.session)

    @property
    def canonical_uri(self) -> Optional[URI]:
        """The absolute location of the (sub)schema.
        
        This is not necessarily an 'absolute URI', as it may contain
        a fragment.
        """
        if self._uri is not None:
            return self._uri

        keys = deque()
        node = self
        while node.parent is not None:
            keys.appendleft(node.key)
            node = node.parent

            if isinstance(node, JSONSchema) and node._uri is not None:
                if fragment := node._uri.fragment:
                    relpath = JSONPointer.parse_uri_fragment(fragment) / keys
                else:
                    relpath = JSONPointer(keys)

                return node._uri.copy(fragment=relpath.uri_fragment())


class Result:
    """The result of evaluating a JSON document
    node against a JSON schema node.

    The root of a :class:`Result` tree represents
    a complete document evaluation result.
    """

    def __init__(
            self,
            schema: JSONSchema,
            instance: JSON,
            *,
            parent: Result = None,
            key: str = None,
    ) -> None:
        self.path: JSONPointer
        """The dynamic evaluation path to the current schema node."""

        self.relpath: JSONPointer
        """The path to the current schema node relative to the evaluating (sub)schema."""

        self.schema: JSONSchema = schema
        """The evaluating (sub)schema."""

        self.instance: JSON = instance
        """The instance under evaluation."""

        self.parent: Optional[Result] = parent
        """The parent result node."""

        self.key: Optional[str] = key
        """The index of the current schema node within its dynamic parent."""

        self.children: Dict[Tuple[str, JSONPointer], Result] = {}
        """Subresults of the current result node, indexed by schema key and instance path."""

        self.annotation: JSONCompatible = None
        """The annotation value of the result."""

        self.error: JSONCompatible = None
        """The error value of the result."""

        self._valid = True
        self._assert = True
        self._discard = False
        self._refschema: Optional[JSONSchema] = None

        if parent is None:
            self.path = JSONPointer()
            self.relpath = JSONPointer()
            self._globals = {}
        else:
            self.path = parent.path / key
            self.relpath = parent.relpath / key if schema is parent.schema else JSONPointer((key,))
            self._globals = None

    @contextmanager
    def __call__(
            self,
            instance: JSON,
            key: str,
            schema: JSONSchema = None,
            *,
            cls: Type[Result] = None,
    ) -> ContextManager[Result]:
        """Yield a subresult for the evaluation of `instance`.
        Descend down the evaluation path by `key`, into `schema` if given, or
        within `self.schema` otherwise.

        Extension keywords may provide a custom :class:`Result` class via `cls`,
        which is applied to all nodes within the yielded subtree.
        """
        if schema is None:
            schema = self.schema

        self.children[key, instance.path] = (child := (cls or self.__class__)(
            schema,
            instance,
            parent=self,
            key=key,
        ))

        try:
            yield child
        finally:
            if child._discard:
                del self.children[key, instance.path]

    @cached_property
    def globals(self) -> Dict:
        root = self
        while root.parent is not None:
            root = root.parent
        return root._globals

    @cached_property
    def schema_node(self) -> JSON:
        """Return the current schema node."""
        return self.relpath.evaluate(self.schema)

    def sibling(self, instance: JSON, key: str) -> Optional[Result]:
        """Return a sibling schema node's evaluation result for `instance`."""
        try:
            return self.parent.children[key, instance.path] if self.parent else None
        except KeyError:
            return None

    def annotate(self, value: JSONCompatible) -> None:
        """Annotate the result."""
        self.annotation = value

    def fail(self, error: JSONCompatible = None) -> None:
        """Mark the result as invalid, optionally with an error."""
        self._valid = False
        self.error = error

    def pass_(self) -> None:
        """Mark the result as valid.
        
        A result is initially valid, so this should only need
        to be called by a keyword when it must reverse a failure.
        """
        self._valid = True
        self.error = None

    def noassert(self) -> None:
        """Indicate that evaluation passes regardless of validity."""
        self._assert = False

    def discard(self) -> None:
        """Indicate that the result should be ignored and discarded."""
        self._discard = True

    def refschema(self, schema: JSONSchema) -> None:
        """Set the referenced schema for a by-reference keyword.
        
        This ensures that :attr:`absolute_uri` returns the URI of the
        referenced schema rather than the referencing keyword.
        """
        self._refschema = schema

    @property
    def valid(self) -> bool:
        """Return the validity of the instance against the schema."""
        return self._valid

    @property
    def passed(self) -> bool:
        """Return the assertion result for the schema node.

        In the standard JSON Schema vocabulary, this can only differ
        from :attr:`valid` for the ``if`` keyword: validity may be false
        (triggering ``else``) while its assertion result is always true.

        For the root result node, :attr:`passed` will always equal :attr:`valid`.
        """
        return self._valid or not self._assert

    @property
    def absolute_uri(self) -> Optional[URI]:
        """Return the absolute URI of the current schema node."""
        if self._refschema is not None:
            return self._refschema.canonical_uri

        if (schema_uri := self.schema.canonical_uri) is not None:
            if fragment := schema_uri.fragment:
                relpath = JSONPointer.parse_uri_fragment(fragment) / self.relpath
            else:
                relpath = self.relpath
            return schema_uri.copy(fragment=relpath.uri_fragment())

    def collect_annotations(self, instance: JSON = None, key: str = None) -> Iterator[JSONCompatible]:
        """Return an iterator over annotations produced in this subtree,
        optionally filtered by instance and/or keyword."""
        if self._valid and not self._discard:
            if self.annotation is not None and \
                    (key is None or key == self.key) and \
                    (instance is None or instance.path == self.instance.path):
                yield self.annotation
            for child in self.children.values():
                yield from child.collect_annotations(instance, key)

    def collect_errors(self, instance: JSON = None, key: str = None) -> Iterator[JSONCompatible]:
        """Return an iterator over errors produced in this subtree,
        optionally filtered by instance and/or keyword."""
        if not self._valid and not self._discard:
            if self.error is not None and \
                    (key is None or key == self.key) and \
                    (instance is None or instance.path == self.instance.path):
                yield self.error
            for child in self.children.values():
                yield from child.collect_errors(instance, key)

    def output(self, format: str, **kwargs: Any) -> JSONCompatible:
        """Return the evaluation result in the specified `format`.

        :param format: One of the standard JSON Schema output formats --
            ``flag``, ``basic``, ``detailed`` or ``verbose`` -- or any
            format registered with the :func:`~jschon.output.output_formatter`
            decorator.
        :param kwargs: Keyword arguments to pass to the output formatter.
        """
        from jschon.output import create_output
        return create_output(self, format, **kwargs)

    def __str__(self) -> str:
        s = 'valid' if self.valid else 'invalid'
        if self.parent:
            s = f'{self.path}: {s}'
        return s



import re

from jschon.json import JSON
from jschon.jsonschema import Result
from jschon.vocabulary import Applicator, ArrayApplicator, Keyword, PropertyApplicator

__all__ = [
    'AllOfKeyword',
    'AnyOfKeyword',
    'OneOfKeyword',
    'NotKeyword',
    'IfKeyword',
    'ThenKeyword',
    'ElseKeyword',
    'DependentSchemasKeyword',
    'PrefixItemsKeyword',
    'ItemsKeyword',
    'UnevaluatedItemsKeyword',
    'ContainsKeyword',
    'PropertiesKeyword',
    'PatternPropertiesKeyword',
    'AdditionalPropertiesKeyword',
    'UnevaluatedPropertiesKeyword',
    'PropertyNamesKeyword',
]


class AllOfKeyword(Keyword, ArrayApplicator):
    key = "allOf"

    def evaluate(self, instance: JSON, result: Result) -> None:
        err_indices = []
        for index, subschema in enumerate(self.json):
            with result(instance, str(index)) as subresult:
                subschema.evaluate(instance, subresult)
                if not subresult.passed:
                    err_indices += [index]

        if err_indices:
            result.fail(f'The instance is invalid against subschemas {err_indices}')


class AnyOfKeyword(Keyword, ArrayApplicator):
    key = "anyOf"

    def evaluate(self, instance: JSON, result: Result) -> None:
        valid = False
        for index, subschema in enumerate(self.json):
            with result(instance, str(index)) as subresult:
                subschema.evaluate(instance, subresult)
                if subresult.passed:
                    valid = True

        if not valid:
            result.fail(f'The instance must be valid against at least one subschema')


class OneOfKeyword(Keyword, ArrayApplicator):
    key = "oneOf"

    def evaluate(self, instance: JSON, result: Result) -> None:
        valid_indices = []
        err_indices = []
        for index, subschema in enumerate(self.json):
            with result(instance, str(index)) as subresult:
                subschema.evaluate(instance, subresult)
                if subresult.passed:
                    valid_indices += [index]
                else:
                    err_indices += [index]

        if len(valid_indices) != 1:
            result.fail('The instance must be valid against exactly one subschema; '
                        f'it is valid against {valid_indices} and invalid against {err_indices}')


class NotKeyword(Keyword, Applicator):
    key = "not"

    def evaluate(self, instance: JSON, result: Result) -> None:
        self.json.evaluate(instance, result)

        if result.passed:
            result.fail('The instance must not be valid against the subschema')
        else:
            result.pass_()


class IfKeyword(Keyword, Applicator):
    key = "if"

    def evaluate(self, instance: JSON, result: Result) -> None:
        self.json.evaluate(instance, result)
        result.noassert()


class ThenKeyword(Keyword, Applicator):
    key = "then"
    depends_on = "if",

    def evaluate(self, instance: JSON, result: Result) -> None:
        if (if_ := result.sibling(instance, "if")) and if_.valid:
            self.json.evaluate(instance, result)
        else:
            result.discard()


class ElseKeyword(Keyword, Applicator):
    key = "else"
    depends_on = "if",

    def evaluate(self, instance: JSON, result: Result) -> None:
        if (if_ := result.sibling(instance, "if")) and not if_.valid:
            self.json.evaluate(instance, result)
        else:
            result.discard()


class DependentSchemasKeyword(Keyword, PropertyApplicator):
    key = "dependentSchemas"
    instance_types = "object",

    def evaluate(self, instance: JSON, result: Result) -> None:
        annotation = []
        err_names = []
        for name, subschema in self.json.items():
            if name in instance:
                with result(instance, name) as subresult:
                    subschema.evaluate(instance, subresult)
                    if subresult.passed:
                        annotation += [name]
                    else:
                        err_names += [name]

        if err_names:
            result.fail(f'Properties {err_names} are invalid against '
                        f'the corresponding "dependentSchemas" subschemas')
        else:
            result.annotate(annotation)


class PrefixItemsKeyword(Keyword, ArrayApplicator):
    key = "prefixItems"
    instance_types = "array",

    def evaluate(self, instance: JSON, result: Result) -> None:
        annotation = None
        error = []
        for index, item in enumerate(instance[:len(self.json)]):
            annotation = index
            with result(item, str(index)) as subresult:
                if not self.json[index].evaluate(item, subresult).passed:
                    error += [index]

        if error:
            result.fail(error)
        elif annotation is not None:
            if annotation == len(instance) - 1:
                annotation = True
            result.annotate(annotation)


class ItemsKeyword(Keyword, Applicator):
    key = "items"
    instance_types = "array",
    depends_on = "prefixItems",

    def evaluate(self, instance: JSON, result: Result) -> None:
        if prefix_items := result.sibling(instance, "prefixItems"):
            start_index = len(prefix_items.schema_node)
        else:
            start_index = 0

        annotation = None
        error = []
        for index, item in enumerate(instance[start_index:], start_index):
            if self.json.evaluate(item, result).passed:
                annotation = True
            else:
                error += [index]
                # reset to passed for the next iteration
                result.pass_()

        if error:
            result.fail(error)
        else:
            result.annotate(annotation)


class UnevaluatedItemsKeyword(Keyword, Applicator):
    key = "unevaluatedItems"
    instance_types = "array",
    depends_on = "prefixItems", "items", "contains", "if", "then", "else", "allOf", "anyOf", "oneOf", "not",

    def evaluate(self, instance: JSON, result: Result) -> None:
        last_evaluated_item = -1
        for prefix_items_annotation in result.parent.collect_annotations(instance, "prefixItems"):
            if prefix_items_annotation is True:
                result.discard()
                return
            if prefix_items_annotation > last_evaluated_item:
                last_evaluated_item = prefix_items_annotation

        for items_annotation in result.parent.collect_annotations(instance, "items"):
            if items_annotation is True:
                result.discard()
                return

        for unevaluated_items_annotation in result.parent.collect_annotations(instance, "unevaluatedItems"):
            if unevaluated_items_annotation is True:
                result.discard()
                return

        contains_indices = set()
        for contains_annotation in result.parent.collect_annotations(instance, "contains"):
            contains_indices |= set(contains_annotation)

        annotation = None
        error = []
        for index, item in enumerate(instance[(start := last_evaluated_item + 1):], start):
            if index not in contains_indices:
                if self.json.evaluate(item, result).passed:
                    annotation = True
                else:
                    error += [index]
                    # reset to passed for the next iteration
                    result.pass_()

        if error:
            result.fail(error)
        else:
            result.annotate(annotation)


class ContainsKeyword(Keyword, Applicator):
    key = "contains"
    instance_types = "array",

    def evaluate(self, instance: JSON, result: Result) -> None:
        annotation = []
        for index, item in enumerate(instance):
            if self.json.evaluate(item, result).passed:
                annotation += [index]
            else:
                result.pass_()

        result.annotate(annotation)
        if not annotation:
            result.fail('The array does not contain any element that is valid '
                        f'against the "{self.key}" subschema')


class PropertiesKeyword(Keyword, PropertyApplicator):
    key = "properties"
    instance_types = "object",

    def evaluate(self, instance: JSON, result: Result) -> None:
        annotation = []
        err_names = []
        for name, item in instance.items():
            if name in self.json:
                with result(item, name) as subresult:
                    self.json[name].evaluate(item, subresult)
                    if subresult.passed:
                        annotation += [name]
                    else:
                        err_names += [name]

        if err_names:
            result.fail(f"Properties {err_names} are invalid")
        else:
            result.annotate(annotation)


class PatternPropertiesKeyword(Keyword, PropertyApplicator):
    key = "patternProperties"
    instance_types = "object",

    def evaluate(self, instance: JSON, result: Result) -> None:
        matched_names = set()
        err_names = []
        for name, item in instance.items():
            for regex, subschema in self.json.items():
                if re.search(regex, name) is not None:
                    with result(item, regex) as subresult:
                        subschema.evaluate(item, subresult)
                        if subresult.passed:
                            matched_names |= {name}
                        else:
                            err_names += [name]

        if err_names:
            result.fail(f"Properties {err_names} are invalid")
        else:
            result.annotate(list(matched_names))


class AdditionalPropertiesKeyword(Keyword, Applicator):
    key = "additionalProperties"
    instance_types = "object",
    depends_on = "properties", "patternProperties",

    def evaluate(self, instance: JSON, result: Result) -> None:
        if properties := result.sibling(instance, "properties"):
            known_property_names = properties.schema_node.keys()
        else:
            known_property_names = ()

        if pattern_properties := result.sibling(instance, "patternProperties"):
            known_property_patterns = pattern_properties.schema_node.keys()
        else:
            known_property_patterns = ()

        annotation = []
        error = []
        for name, item in instance.items():
            if name not in known_property_names and not any(
                    re.search(regex, name) for regex in known_property_patterns
            ):
                if self.json.evaluate(item, result).passed:
                    annotation += [name]
                else:
                    error += [name]
                    # reset to passed for the next iteration
                    result.pass_()

        if error:
            result.fail(error)
        else:
            result.annotate(annotation)


class UnevaluatedPropertiesKeyword(Keyword, Applicator):
    key = "unevaluatedProperties"
    instance_types = "object",
    depends_on = "properties", "patternProperties", "additionalProperties", \
                 "if", "then", "else", "dependentSchemas", \
                 "allOf", "anyOf", "oneOf", "not",

    def evaluate(self, instance: JSON, result: Result) -> None:
        evaluated_names = set()
        for properties_annotation in result.parent.collect_annotations(instance, "properties"):
            evaluated_names |= set(properties_annotation)
        for pattern_properties_annotation in result.parent.collect_annotations(instance, "patternProperties"):
            evaluated_names |= set(pattern_properties_annotation)
        for additional_properties_annotation in result.parent.collect_annotations(instance, "additionalProperties"):
            evaluated_names |= set(additional_properties_annotation)
        for unevaluated_properties_annotation in result.parent.collect_annotations(instance, "unevaluatedProperties"):
            evaluated_names |= set(unevaluated_properties_annotation)

        annotation = []
        error = []
        for name, item in instance.items():
            if name not in evaluated_names:
                if self.json.evaluate(item, result).passed:
                    annotation += [name]
                else:
                    error += [name]
                    # reset to passed for the next iteration
                    result.pass_()

        if error:
            result.fail(error)
        else:
            result.annotate(annotation)


class PropertyNamesKeyword(Keyword, Applicator):
    key = "propertyNames"
    instance_types = "object",

    def evaluate(self, instance: JSON, result: Result) -> None:
        error = []
        for name in instance:
            if not self.json.evaluate(JSON(name, parent=instance, key=name), result).passed:
                error += [name]
                result.pass_()

        if error:
            result.fail(error)



from jschon.json import JSON
from jschon.jsonschema import Result
from jschon.vocabulary import Keyword

__all__ = [
    'TitleKeyword',
    'DescriptionKeyword',
    'DefaultKeyword',
    'DeprecatedKeyword',
    'ReadOnlyKeyword',
    'WriteOnlyKeyword',
    'ExamplesKeyword',
    'ContentMediaTypeKeyword',
    'ContentEncodingKeyword',
    'ContentSchemaKeyword',
]


class AnnotationKeyword(Keyword):

    def evaluate(self, instance: JSON, result: Result) -> None:
        result.annotate(self.json.data)
        result.noassert()


class TitleKeyword(AnnotationKeyword):
    key = "title"


class DescriptionKeyword(AnnotationKeyword):
    key = "description"


class DefaultKeyword(AnnotationKeyword):
    key = "default"


class DeprecatedKeyword(AnnotationKeyword):
    key = "deprecated"


class ReadOnlyKeyword(AnnotationKeyword):
    key = "readOnly"


class WriteOnlyKeyword(AnnotationKeyword):
    key = "writeOnly"


class ExamplesKeyword(AnnotationKeyword):
    key = "examples"


class ContentMediaTypeKeyword(AnnotationKeyword):
    key = "contentMediaType"
    instance_types = "string",


class ContentEncodingKeyword(AnnotationKeyword):
    key = "contentEncoding"
    instance_types = "string",


class ContentSchemaKeyword(AnnotationKeyword):
    key = "contentSchema"
    instance_types = "string",
    depends_on = "contentMediaType",

    def evaluate(self, instance: JSON, result: Result) -> None:
        if result.sibling(instance, "contentMediaType"):
            super().evaluate(instance, result)
        else:
            result.discard()



import json
from os import PathLike
from typing import Any, Tuple, Iterable, Union

__all__ = [
    'tuplify',
    'json_dumpf',
    'json_dumps',
    'json_loadf',
    'json_loadr',
    'json_loads',
]

requests = None


def _import_requests():
    global requests
    try:
        import requests
    except ImportError as e:
        raise ImportError('requests is not installed, run `pip install jschon[requests]`') from e


def tuplify(value: Any) -> Tuple:
    if value is None:
        return ()
    if isinstance(value, Tuple):
        return value
    if isinstance(value, Iterable) and not isinstance(value, str):
        return tuple(value)
    return value,


def json_dumpf(obj: Any, path: Union[str, PathLike]) -> None:
    """Serialize a JSON-compatible Python object to a JSON file."""
    with open(path, 'w') as f:
        json.dump(obj, f, ensure_ascii=False, allow_nan=False, default=_serialize_custom_obj)


def json_dumps(obj: Any) -> str:
    """Serialize a JSON-compatible Python object to a JSON string."""
    return json.dumps(obj, ensure_ascii=False, allow_nan=False, default=_serialize_custom_obj)


def json_loadf(path: Union[str, PathLike]) -> Any:
    """Deserialize a JSON file, returning a JSON-compatible Python object."""
    with open(path) as f:
        return json.load(f, parse_constant=_parse_invalid_const)


def json_loadr(url: str) -> Any:
    """Fetch and deserialize a remote JSON resource, returning a JSON-compatible Python object."""
    if requests is None:
        _import_requests()
    r = requests.get(url)
    r.raise_for_status()
    return r.json(parse_constant=_parse_invalid_const)


def json_loads(value: str) -> Any:
    """Deserialize a JSON string, returning a JSON-compatible Python object."""
    return json.loads(value, parse_constant=_parse_invalid_const)


def _parse_invalid_const(c):
    """Called when '-Infinity', 'Infinity' or 'NaN' is encountered in the
    JSON-encoded input. These JavaScript constants are not strictly allowed
    by the JSON data model."""
    raise ValueError(f"{c} is not a valid JSON value")


def _serialize_custom_obj(o):
    from jschon.json import JSON
    if isinstance(o, JSON):
        return o.data
    raise TypeError



from __future__ import annotations

import rfc3986
import rfc3986.exceptions
import rfc3986.misc
import rfc3986.validators

from jschon.exceptions import URIError

__all__ = [
    'URI',
]


class URI:
    def __init__(self, value: str) -> None:
        self._uriref = rfc3986.uri_reference(value)

    def __str__(self) -> str:
        return self._uriref.unsplit()

    def __repr__(self) -> str:
        return f"URI({str(self)!r})"

    def __len__(self) -> int:
        return len(str(self))

    def __hash__(self) -> int:
        return hash(self._uriref)

    def __eq__(self, other) -> bool:
        if isinstance(other, URI):
            return self._uriref == other._uriref
        if other is None:
            return False
        return self._uriref.__eq__(other)

    @property
    def scheme(self) -> str:
        return self._uriref.scheme

    @property
    def authority(self) -> str:
        return self._uriref.authority

    @property
    def path(self) -> str:
        return self._uriref.path

    @property
    def query(self) -> str:
        return self._uriref.query

    @property
    def fragment(self) -> str:
        return self._uriref.fragment

    def is_absolute(self) -> bool:
        return self._uriref.is_absolute()

    def has_absolute_base(self) -> bool:
        return self.copy(fragment=False).is_absolute()

    def resolve(self, base_uri: URI) -> URI:
        """Produce a new URI by resolving self against the given base URI."""
        uri = object.__new__(URI)
        uri._uriref = self._uriref.resolve_with(base_uri._uriref)
        return uri

    def copy(
            self,
            scheme=True,
            authority=True,
            path=True,
            query=True,
            fragment=True,
    ) -> URI:
        """Produce a new URI composed of the specified components of self.

        - True => use existing
        - False/None => remove
        - Otherwise => replace
        """
        uri = object.__new__(URI)
        uri._uriref = self._uriref.copy_with(
            scheme=rfc3986.misc.UseExisting if scheme is True else None if scheme is False else scheme,
            authority=rfc3986.misc.UseExisting if authority is True else None if authority is False else authority,
            path=rfc3986.misc.UseExisting if path is True else None if path is False else path,
            query=rfc3986.misc.UseExisting if query is True else None if query is False else query,
            fragment=rfc3986.misc.UseExisting if fragment is True else None if fragment is False else fragment,
        )
        return uri

    def validate(
            self,
            require_scheme: bool = False,
            require_normalized: bool = False,
            allow_fragment: bool = True,
            allow_non_empty_fragment: bool = True,
    ) -> None:
        """Validate self.

        :raise URIError: if self fails validation
        """
        validator = rfc3986.validators.Validator()
        if require_scheme:
            validator = validator.require_presence_of('scheme')
        try:
            validator.validate(self._uriref)
        except rfc3986.exceptions.ValidationError as e:
            msg = f"'{self}' is not a valid URI"
            if require_scheme:
                msg += " or does not contain a scheme"
            raise URIError(msg) from e

        if require_normalized and self._uriref != self._uriref.normalize():
            raise URIError(f"'{self}' is not normalized")

        if not allow_fragment and self._uriref.fragment is not None:
            raise URIError(f"'{self}' has a fragment")

        if not allow_non_empty_fragment and self._uriref.fragment:
            raise URIError(f"'{self}' has a non-empty fragment")



from __future__ import annotations

import pathlib
import uuid
from contextlib import contextmanager
from os import PathLike
from typing import Any, ContextManager, Dict, Hashable, Mapping, Union

from jschon.exceptions import CatalogError, JSONPointerError, URIError
from jschon.json import JSONCompatible
from jschon.jsonpointer import JSONPointer
from jschon.jsonschema import JSONSchema
from jschon.uri import URI
from jschon.utils import json_loadf, json_loadr
from jschon.vocabulary import KeywordClass, Metaschema, Vocabulary
from jschon.vocabulary.format import FormatValidator

__all__ = [
    'Catalog',
    'Source',
    'LocalSource',
    'RemoteSource',
]


class Source:
    def __init__(self, suffix: str = None) -> None:
        self.suffix = suffix

    def __call__(self, relative_path: str) -> JSONCompatible:
        raise NotImplementedError


class LocalSource(Source):
    def __init__(self, base_dir: Union[str, PathLike], **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.base_dir = base_dir

    def __call__(self, relative_path: str) -> JSONCompatible:
        filepath = pathlib.Path(self.base_dir) / relative_path
        if self.suffix:
            filepath = str(filepath)
            filepath += self.suffix

        return json_loadf(filepath)


class RemoteSource(Source):
    def __init__(self, base_url: URI, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.base_url = base_url

    def __call__(self, relative_path: str) -> JSONCompatible:
        url = str(URI(relative_path).resolve(self.base_url))
        if self.suffix:
            url += self.suffix

        return json_loadr(url)


class Catalog:
    """The :class:`Catalog` acts as a schema cache, enabling schemas and subschemas
    to be indexed, re-used, and cross-referenced by URI. The cache is transparently
    partitioned by (arbitrary) session identifiers, which may optionally be provided
    when creating :class:`~jschon.jsonschema.JSONSchema` objects."""

    _catalog_registry: Dict[Hashable, Catalog] = {}

    @classmethod
    def get_catalog(cls, name: str = 'catalog') -> Catalog:
        try:
            return cls._catalog_registry[name]
        except KeyError:
            raise CatalogError(f'Catalog name "{name}" not found.')

    def __init__(self, name: str = 'catalog') -> None:
        """Initialize a :class:`Catalog` instance.

        :param name: a unique name for this :class:`Catalog` instance
        """
        self.__class__._catalog_registry[name] = self

        self.name: str = name
        """The unique name of this :class:`Catalog` instance."""

        self._uri_sources: Dict[URI, Source] = {}
        self._vocabularies: Dict[URI, Vocabulary] = {}
        self._format_validators: Dict[str, FormatValidator] = {}
        self._schema_cache: Dict[Hashable, Dict[URI, JSONSchema]] = {}

    def add_uri_source(self, base_uri: URI, source: Source):
        """Register a source for URI-identified JSON resources.

        :param base_uri: a normalized, absolute URI - including scheme, without
            a fragment, and ending with ``'/'``
        :param source: a :class:`Source` object
        :raise CatalogError: if `base_uri` is invalid
        """
        try:
            base_uri.validate(require_scheme=True, require_normalized=True, allow_fragment=False)
        except URIError as e:
            raise CatalogError from e

        if not base_uri.path or not base_uri.path.endswith('/'):
            raise CatalogError('base_uri must end with "/"')

        self._uri_sources[base_uri] = source

    def load_json(self, uri: URI) -> JSONCompatible:
        """Load a JSON-compatible object from the source for `uri`.

        If there are multiple candidate base URIs for `uri`, the most specific
        match (i.e. the longest one) is selected.

        :param uri: a normalized, absolute URI - including scheme, without
            a fragment
        :raise CatalogError: if `uri` is invalid, a source is not available
            for `uri`, or if a loading error occurs
        """
        try:
            uri.validate(require_scheme=True, require_normalized=True, allow_fragment=False)
        except URIError as e:
            raise CatalogError from e

        uristr = str(uri)
        candidates = [
            (base_uristr, source)
            for base_uri, source in self._uri_sources.items()
            if uristr.startswith(base_uristr := str(base_uri))
        ]
        if candidates:
            candidates.sort(key=lambda c: len(c[0]), reverse=True)
            base_uristr, source = candidates[0]
            relative_path = uristr[len(base_uristr):]
            try:
                return source(relative_path)
            except Exception as e:
                raise CatalogError(*e.args) from e

        raise CatalogError(f'A source is not available for "{uri}"')

    def create_vocabulary(self, uri: URI, *kwclasses: KeywordClass) -> None:
        """Create a :class:`~jschon.vocabulary.Vocabulary` object, which
        may be used by a :class:`~jschon.vocabulary.Metaschema` to provide
        keyword classes used in schema construction.

        :param uri: the URI identifying the vocabulary
        :param kwclasses: the :class:`~jschon.vocabulary.Keyword` classes
            constituting the vocabulary
        """
        self._vocabularies[uri] = Vocabulary(uri, *kwclasses)

    def get_vocabulary(self, uri: URI) -> Vocabulary:
        """Get a :class:`~jschon.vocabulary.Vocabulary` by its `uri`.

        :param uri: the URI identifying the vocabulary
        :raise CatalogError: if `uri` is not a recognized vocabulary URI
        """
        try:
            return self._vocabularies[uri]
        except KeyError:
            raise CatalogError(f"Unrecognized vocabulary URI '{uri}'")

    def create_metaschema(
            self,
            uri: URI,
            core_vocabulary_uri: URI,
            *default_vocabulary_uris: URI,
            **kwargs: Any,
    ) -> None:
        """Create, cache and validate a :class:`~jschon.vocabulary.Metaschema`.

        :param uri: the URI identifying the metaschema
        :param core_vocabulary_uri: the URI identifying the metaschema's
            core :class:`~jschon.vocabulary.Vocabulary`
        :param default_vocabulary_uris: default :class:`~jschon.vocabulary.Vocabulary`
            URIs, used in the absence of a ``"$vocabulary"`` keyword in the
            metaschema JSON file
        :param kwargs: additional keyword arguments to pass through to the
            :class:`~jschon.jsonschema.JSONSchema` constructor
        """
        metaschema_doc = self.load_json(uri)
        core_vocabulary = self.get_vocabulary(core_vocabulary_uri)
        default_vocabularies = [
            self.get_vocabulary(vocab_uri)
            for vocab_uri in default_vocabulary_uris
        ]
        metaschema = Metaschema(
            self,
            metaschema_doc,
            core_vocabulary,
            *default_vocabularies,
            **kwargs,
            uri=uri,
        )
        if not metaschema.validate().valid:
            raise CatalogError("The metaschema is invalid against itself")

    def add_format_validators(self, validators: Mapping[str, FormatValidator]) -> None:
        """Register a collection of format validators.
        
        In jschon, a given occurrence of the ``"format"`` keyword evaluates
        a JSON instance using a format validation callable, if one has been
        registered for the applicable *format attribute* (the keyword's value).
        If a validator has not been registered for that format attribute,
        keyword evaluation simply passes.

        :param validators: a dictionary of :class:`~jschon.vocabulary.format.FormatValidator`
            callables, keyed by format attribute
        """
        self._format_validators.update(validators)

    def get_format_validator(self, format_attr: str) -> FormatValidator:
        """Get a registered :class:`~jschon.vocabulary.format.FormatValidator`
        function.

        :param format_attr: the format attribute (``"format"`` keyword value)
            to which the validator applies
        :raise CatalogError: if no format validator is registered for the
            given `format_attr`
        """
        try:
            return self._format_validators[format_attr]
        except KeyError:
            raise CatalogError(f"Unsupported format attribute '{format_attr}'")

    def add_schema(
            self,
            uri: URI,
            schema: JSONSchema,
            *,
            session: Hashable = 'default',
    ) -> None:
        """Add a (sub)schema to a session cache.

        :param uri: the URI identifying the (sub)schema
        :param schema: the :class:`~jschon.jsonschema.JSONSchema` instance to cache
        :param session: a session identifier
        """
        self._schema_cache.setdefault(session, {})
        self._schema_cache[session][uri] = schema

    def del_schema(
            self,
            uri: URI,
            *,
            session: Hashable = 'default',
    ) -> None:
        """Remove a (sub)schema from a session cache.

        :param uri: the URI identifying the (sub)schema
        :param session: a session identifier
        """
        if session in self._schema_cache:
            self._schema_cache[session].pop(uri, None)

    def get_schema(
            self,
            uri: URI,
            *,
            metaschema_uri: URI = None,
            session: Hashable = 'default',
    ) -> JSONSchema:
        """Get a (sub)schema identified by `uri` from a session cache, or
        load it from disk if not already cached.

        :param uri: the URI identifying the (sub)schema
        :param metaschema_uri: passed to the :class:`~jschon.jsonschema.JSONSchema`
            constructor when loading a new instance from disk
        :param session: a session identifier
        :raise CatalogError: if a schema cannot be found for `uri`, or if the
            object referenced by `uri` is not a :class:`~jschon.jsonschema.JSONSchema`
        """
        try_caches = ('__meta__',) \
            if session == '__meta__' \
            else (session, '__meta__')

        for cacheid in try_caches:
            try:
                return self._schema_cache[cacheid][uri]
            except KeyError:
                pass

        schema = None
        base_uri = uri.copy(fragment=False)

        if uri.fragment is not None:
            for cacheid in try_caches:
                try:
                    schema = self._schema_cache[cacheid][base_uri]
                    break
                except KeyError:
                    pass

        if schema is None:
            doc = self.load_json(base_uri)
            schema = JSONSchema(
                doc,
                catalog=self,
                session=session,
                uri=base_uri,
                metaschema_uri=metaschema_uri,
            )

        if uri.fragment:
            try:
                ptr = JSONPointer.parse_uri_fragment(uri.fragment)
                schema = ptr.evaluate(schema)
            except JSONPointerError as e:
                raise CatalogError(f"Schema not found for {uri}") from e

        if not isinstance(schema, JSONSchema):
            raise CatalogError(f"The object referenced by {uri} is not a JSON Schema")

        return schema

    @contextmanager
    def session(self, session: Hashable = None) -> ContextManager[Hashable]:
        if session is None:
            session = uuid.uuid4()

        if session in self._schema_cache:
            raise CatalogError("session is already in use")

        try:
            yield session
        finally:
            self._schema_cache.pop(session, None)



from typing import Callable

from jschon.exceptions import CatalogError
from jschon.json import JSON, JSONCompatible
from jschon.jsonschema import JSONSchema, Result
from jschon.vocabulary import Keyword

__all__ = [
    'FormatKeyword',
    'FormatValidator',
]

FormatValidator = Callable[[JSONCompatible], None]
"""The type of a ``"format"`` keyword validator.

A :class:`~jschon.vocabulary.format.FormatValidator` is a
callable accepting a JSON-compatible Python object as its
only argument. It must raise a :exc:`ValueError` if the
argument is invalid per the applicable format specification.
"""


class FormatKeyword(Keyword):
    key = "format"

    def __init__(self, parentschema: JSONSchema, value: str):
        super().__init__(parentschema, value)

        try:
            self.validator: FormatValidator = parentschema.catalog.get_format_validator(value)
        except CatalogError:
            self.validator = None

    def evaluate(self, instance: JSON, result: Result) -> None:
        result.annotate(self.json.data)
        if self.validator is not None:
            try:
                self.validator(instance.data)
            except ValueError as e:
                result.fail(f'The instance is invalid against the "{self.json.data}" format: {e}')
        else:
            result.noassert()



import re
from decimal import Decimal, InvalidOperation

from jschon.json import JSON
from jschon.jsonschema import JSONSchema, Result
from jschon.utils import tuplify
from jschon.vocabulary import Keyword

__all__ = [
    'TypeKeyword',
    'EnumKeyword',
    'ConstKeyword',
    'MultipleOfKeyword',
    'MaximumKeyword',
    'ExclusiveMaximumKeyword',
    'MinimumKeyword',
    'ExclusiveMinimumKeyword',
    'MaxLengthKeyword',
    'MinLengthKeyword',
    'PatternKeyword',
    'MaxItemsKeyword',
    'MinItemsKeyword',
    'UniqueItemsKeyword',
    'MaxContainsKeyword',
    'MinContainsKeyword',
    'MaxPropertiesKeyword',
    'MinPropertiesKeyword',
    'RequiredKeyword',
    'DependentRequiredKeyword',
]


class TypeKeyword(Keyword):
    key = "type"

    def evaluate(self, instance: JSON, result: Result) -> None:
        types = tuplify(self.json.data)
        if instance.type in types:
            valid = True
        elif instance.type == "number" and "integer" in types:
            valid = instance.data == int(instance.data)
        else:
            valid = False

        if not valid:
            result.fail(f"The instance must be of type {self.json}")


class EnumKeyword(Keyword):
    key = "enum"

    def evaluate(self, instance: JSON, result: Result) -> None:
        if instance not in self.json:
            result.fail("The instance value must be equal to one of the elements "
                        "in the defined enumeration")


class ConstKeyword(Keyword):
    key = "const"

    def evaluate(self, instance: JSON, result: Result) -> None:
        if instance != self.json:
            result.fail(f"The instance value must be equal to the defined constant")


class MultipleOfKeyword(Keyword):
    key = "multipleOf"
    instance_types = "number",

    def evaluate(self, instance: JSON, result: Result) -> None:
        try:
            if Decimal(f'{instance.data}') % Decimal(f'{self.json.data}') != 0:
                result.fail(f"The value must be a multiple of {self.json}")
        except InvalidOperation:
            result.fail(f"Invalid operation: {instance} % {self.json}")


class MaximumKeyword(Keyword):
    key = "maximum"
    instance_types = "number",

    def evaluate(self, instance: JSON, result: Result) -> None:
        if instance > self.json:
            result.fail(f"The value may not be greater than {self.json}")


class ExclusiveMaximumKeyword(Keyword):
    key = "exclusiveMaximum"
    instance_types = "number",

    def evaluate(self, instance: JSON, result: Result) -> None:
        if instance >= self.json:
            result.fail(f"The value must be less than {self.json}")


class MinimumKeyword(Keyword):
    key = "minimum"
    instance_types = "number",

    def evaluate(self, instance: JSON, result: Result) -> None:
        if instance < self.json:
            result.fail(f"The value may not be less than {self.json}")


class ExclusiveMinimumKeyword(Keyword):
    key = "exclusiveMinimum"
    instance_types = "number",

    def evaluate(self, instance: JSON, result: Result) -> None:
        if instance <= self.json:
            result.fail(f"The value must be greater than {self.json}")


class MaxLengthKeyword(Keyword):
    key = "maxLength"
    instance_types = "string",

    def evaluate(self, instance: JSON, result: Result) -> None:
        if len(instance) > self.json:
            result.fail(f"The text is too long (maximum {self.json} characters)")


class MinLengthKeyword(Keyword):
    key = "minLength"
    instance_types = "string",

    def evaluate(self, instance: JSON, result: Result) -> None:
        if len(instance) < self.json:
            result.fail(f"The text is too short (minimum {self.json} characters)")


class PatternKeyword(Keyword):
    key = "pattern"
    instance_types = "string",

    def __init__(self, parentschema: JSONSchema, value: str):
        super().__init__(parentschema, value)
        self.regex = re.compile(value)

    def evaluate(self, instance: JSON, result: Result) -> None:
        if self.regex.search(instance.data) is None:
            result.fail(f"The text must match the regular expression {self.json}")


class MaxItemsKeyword(Keyword):
    key = "maxItems"
    instance_types = "array",

    def evaluate(self, instance: JSON, result: Result) -> None:
        if len(instance) > self.json:
            result.fail(f"The array has too many elements (maximum {self.json})")


class MinItemsKeyword(Keyword):
    key = "minItems"
    instance_types = "array",

    def evaluate(self, instance: JSON, result: Result) -> None:
        if len(instance) < self.json:
            result.fail(f"The array has too few elements (minimum {self.json})")


class UniqueItemsKeyword(Keyword):
    key = "uniqueItems"
    instance_types = "array",

    def evaluate(self, instance: JSON, result: Result) -> None:
        if not self.json.data:
            return

        uniquified = []
        for item in instance:
            if item not in uniquified:
                uniquified += [item]

        if len(instance) > len(uniquified):
            result.fail("The array's elements must all be unique")


class MaxContainsKeyword(Keyword):
    key = "maxContains"
    instance_types = "array",
    depends_on = "contains",

    def evaluate(self, instance: JSON, result: Result) -> None:
        if contains := result.sibling(instance, "contains"):
            if contains.annotation is not None and len(contains.annotation) > self.json:
                result.fail('The array has too many elements matching the '
                            f'"contains" subschema (maximum {self.json})')


class MinContainsKeyword(Keyword):
    key = "minContains"
    instance_types = "array",
    depends_on = "contains", "maxContains",

    def evaluate(self, instance: JSON, result: Result) -> None:
        if contains := result.sibling(instance, "contains"):
            contains_count = len(contains.annotation) \
                if contains.annotation is not None \
                else 0

            valid = contains_count >= self.json

            if valid and not contains.valid:
                max_contains = result.sibling(instance, "maxContains")
                if not max_contains or max_contains.valid:
                    contains.pass_()

            if not valid:
                result.fail('The array has too few elements matching the '
                            f'"contains" subschema (minimum {self.json})')


class MaxPropertiesKeyword(Keyword):
    key = "maxProperties"
    instance_types = "object",

    def evaluate(self, instance: JSON, result: Result) -> None:
        if len(instance) > self.json:
            result.fail(f"The object has too many properties (maximum {self.json})")


class MinPropertiesKeyword(Keyword):
    key = "minProperties"
    instance_types = "object",

    def evaluate(self, instance: JSON, result: Result) -> None:
        if len(instance) < self.json:
            result.fail(f"The object has too few properties (minimum {self.json})")


class RequiredKeyword(Keyword):
    key = "required"
    instance_types = "object",

    def evaluate(self, instance: JSON, result: Result) -> None:
        missing = [name.value for name in self.json if name.data not in instance]
        if missing:
            result.fail(f"The object is missing required properties {missing}")


class DependentRequiredKeyword(Keyword):
    key = "dependentRequired"
    instance_types = "object",

    def evaluate(self, instance: JSON, result: Result) -> None:
        missing = {}
        for name, dependents in self.json.items():
            if name in instance:
                missing_deps = [dep for dep in dependents if dep.data not in instance]
                if missing_deps:
                    missing[name] = missing_deps

        if missing:
            result.fail(f"The object is missing dependent properties {missing}")



from .catalog import Catalog, LocalSource, RemoteSource
from .exceptions import CatalogError, JSONPatchError, JSONPointerError, JSONSchemaError, RelativeJSONPointerError, URIError
from .json import JSON
from .jsonpatch import JSONPatch, JSONPatchOperation
from .jsonpointer import JSONPointer, RelativeJSONPointer
from .jsonschema import JSONSchema
from .uri import URI

__all__ = [
    'Catalog',
    'CatalogError',
    'JSON',
    'JSONPatch',
    'JSONPatchError',
    'JSONPatchOperation',
    'JSONPointer',
    'JSONPointerError',
    'RelativeJSONPointer',
    'RelativeJSONPointerError',
    'JSONSchema',
    'JSONSchemaError',
    'LocalSource',
    'RemoteSource',
    'URI',
    'URIError',
    'create_catalog',
]

__version__ = '0.9.0'


def create_catalog(*vocabularies: str, name: str = 'catalog') -> Catalog:
    """Create and return a :class:`Catalog` instance, configured with
    support for the specified JSON Schema vocabularies.

    :param vocabularies: any of ``2019-09``, ``2020-12``
    :param name: a unique name for the :class:`Catalog` instance
    :raise ValueError: if a supplied vocabulary parameter is not recognized
    """
    from .catalog import _2019_09, _2020_12

    catalog = Catalog(name=name)

    vocabulary_initializers = {
        '2019-09': _2019_09.initialize,
        '2020-12': _2020_12.initialize,
    }
    try:
        for vocabulary in vocabularies:
            vocabulary_init = vocabulary_initializers[vocabulary]
            vocabulary_init(catalog)

    except KeyError as e:
        raise ValueError(f'Unsupported vocabulary "{e.args[0]}"')

    return catalog



class CatalogError(Exception):
    pass


class JSONPatchError(Exception):
    pass


class JSONPointerError(Exception):
    pass


class RelativeJSONPointerError(Exception):
    pass


class JSONSchemaError(Exception):
    pass


class URIError(Exception):
    pass



from typing import Mapping

from jschon.exceptions import CatalogError, JSONSchemaError, URIError
from jschon.json import JSON
from jschon.jsonschema import JSONSchema, Result
from jschon.uri import URI
from jschon.vocabulary import Keyword, Metaschema, PropertyApplicator

__all__ = [
    'SchemaKeyword',
    'VocabularyKeyword',
    'IdKeyword',
    'RefKeyword',
    'AnchorKeyword',
    'DynamicRefKeyword',
    'DynamicAnchorKeyword',
    'DefsKeyword',
    'CommentKeyword',
]


class SchemaKeyword(Keyword):
    key = "$schema"
    static = True

    def __init__(self, parentschema: JSONSchema, value: str):
        super().__init__(parentschema, value)

        try:
            (uri := URI(value)).validate(require_scheme=True, require_normalized=True)
        except URIError as e:
            raise JSONSchemaError from e

        parentschema.metaschema_uri = uri


class VocabularyKeyword(Keyword):
    key = "$vocabulary"
    static = True

    def __init__(self, parentschema: JSONSchema, value: Mapping[str, bool]):
        super().__init__(parentschema, value)

        if not isinstance(parentschema, Metaschema):
            return

        if (core_vocab_uri := str(parentschema.core_vocabulary.uri)) not in value or \
                value[core_vocab_uri] is not True:
            raise JSONSchemaError(f'The "$vocabulary" keyword must list the core vocabulary with a value of true')

        for vocab_uri, vocab_required in value.items():
            try:
                (vocab_uri := URI(vocab_uri)).validate(require_scheme=True, require_normalized=True)
            except URIError as e:
                raise JSONSchemaError from e

            try:
                vocabulary = parentschema.catalog.get_vocabulary(vocab_uri)
                parentschema.kwclasses.update(vocabulary.kwclasses)
            except CatalogError:
                if vocab_required:
                    raise JSONSchemaError(f"The metaschema requires an unrecognized vocabulary '{vocab_uri}'")


class IdKeyword(Keyword):
    key = "$id"
    static = True

    def __init__(self, parentschema: JSONSchema, value: str):
        super().__init__(parentschema, value)

        (uri := URI(value)).validate(require_normalized=True, allow_fragment=False)
        if not uri.is_absolute():
            if (base_uri := parentschema.base_uri) is not None:
                uri = uri.resolve(base_uri)
            else:
                raise JSONSchemaError(f'No base URI against which to resolve the "$id" value "{value}"')

        parentschema.uri = uri


class RefKeyword(Keyword):
    key = "$ref"

    def __init__(self, parentschema: JSONSchema, value: str):
        super().__init__(parentschema, value)
        self.refschema = None

    def resolve(self) -> None:
        uri = URI(self.json.data)
        if not uri.has_absolute_base():
            if (base_uri := self.parentschema.base_uri) is not None:
                uri = uri.resolve(base_uri)
            else:
                raise JSONSchemaError(f'No base URI against which to resolve the "$ref" value "{uri}"')

        self.refschema = self.parentschema.catalog.get_schema(
            uri, metaschema_uri=self.parentschema.metaschema_uri, session=self.parentschema.session
        )

    def evaluate(self, instance: JSON, result: Result) -> None:
        self.refschema.evaluate(instance, result)
        result.refschema(self.refschema)


class AnchorKeyword(Keyword):
    key = "$anchor"
    static = True

    def __init__(self, parentschema: JSONSchema, value: str):
        super().__init__(parentschema, value)

        if (base_uri := parentschema.base_uri) is not None:
            uri = URI(f'{base_uri}#{value}')
        else:
            raise JSONSchemaError(f'No base URI for "$anchor" value "{value}"')

        parentschema.catalog.add_schema(uri, parentschema, session=parentschema.session)


class DynamicRefKeyword(Keyword):
    key = "$dynamicRef"

    def __init__(self, parentschema: JSONSchema, value: str):
        super().__init__(parentschema, value)

        # this is not required by the spec, but it doesn't make sense
        # for a $dynamicRef *not* to end in a plain-name fragment
        if (fragment := URI(value).fragment) is None or '/' in fragment:
            raise JSONSchemaError('The value for "$dynamicRef" must end in a plain-name fragment')

        self.fragment = fragment
        self.refschema = None
        self.dynamic = False

    def resolve(self) -> None:
        uri = URI(self.json.data)
        if not uri.has_absolute_base():
            if (base_uri := self.parentschema.base_uri) is not None:
                uri = uri.resolve(base_uri)
            else:
                raise JSONSchemaError(f'No base URI against which to resolve the "$dynamicRef" value "{uri}"')

        self.refschema = self.parentschema.catalog.get_schema(
            uri, metaschema_uri=self.parentschema.metaschema_uri, session=self.parentschema.session
        )
        if (dynamic_anchor := self.refschema.get("$dynamicAnchor")) and dynamic_anchor.data == self.fragment:
            self.dynamic = True

    def evaluate(self, instance: JSON, result: Result) -> None:
        refschema = self.refschema

        if self.dynamic:
            target = result
            checked_uris = set()

            while target is not None:
                if (base_uri := target.schema.base_uri) is not None and base_uri not in checked_uris:
                    checked_uris |= {base_uri}
                    target_uri = URI(f"#{self.fragment}").resolve(base_uri)
                    try:
                        found_schema = self.parentschema.catalog.get_schema(
                            target_uri, session=self.parentschema.session
                        )
                        if (dynamic_anchor := found_schema.get("$dynamicAnchor")) and \
                                dynamic_anchor.data == self.fragment:
                            refschema = found_schema
                    except CatalogError:
                        pass

                target = target.parent

        refschema.evaluate(instance, result)
        result.refschema(refschema)


class DynamicAnchorKeyword(Keyword):
    key = "$dynamicAnchor"
    static = True

    def __init__(self, parentschema: JSONSchema, value: str):
        super().__init__(parentschema, value)

        if (base_uri := parentschema.base_uri) is not None:
            uri = URI(f'{base_uri}#{value}')
        else:
            raise JSONSchemaError(f'No base URI for "$dynamicAnchor" value "{value}"')

        parentschema.catalog.add_schema(uri, parentschema, session=parentschema.session)


class DefsKeyword(Keyword, PropertyApplicator):
    key = "$defs"
    static = True


class CommentKeyword(Keyword):
    key = "$comment"
    static = True



from jschon.exceptions import JSONSchemaError
from jschon.json import JSON
from jschon.jsonschema import JSONSchema, Result
from jschon.vocabulary import Applicator, ArrayApplicator, Keyword

__all__ = [
    'RecursiveRefKeyword_2019_09',
    'RecursiveAnchorKeyword_2019_09',
    'ItemsKeyword_2019_09',
    'AdditionalItemsKeyword_2019_09',
    'UnevaluatedItemsKeyword_2019_09',
]


class RecursiveRefKeyword_2019_09(Keyword):
    key = "$recursiveRef"

    def __init__(self, parentschema: JSONSchema, value: str):
        super().__init__(parentschema, value)
        if value != '#':
            raise JSONSchemaError(f'"$recursiveRef" may only take the value "#"')

        self.refschema = None

    def resolve(self) -> None:
        if (base_uri := self.parentschema.base_uri) is not None:
            self.refschema = self.parentschema.catalog.get_schema(
                base_uri, metaschema_uri=self.parentschema.metaschema_uri, session=self.parentschema.session
            )
        else:
            raise JSONSchemaError(f'No base URI against which to resolve "$recursiveRef"')

    def evaluate(self, instance: JSON, result: Result) -> None:
        refschema = self.refschema
        if (recursive_anchor := refschema.get("$recursiveAnchor")) and \
                recursive_anchor.data is True:

            target = result
            while target is not None:
                if (base_anchor := target.schema.get("$recursiveAnchor")) and \
                        base_anchor.data is True:
                    refschema = target.schema

                target = target.parent

        refschema.evaluate(instance, result)
        result.refschema(refschema)


class RecursiveAnchorKeyword_2019_09(Keyword):
    key = "$recursiveAnchor"
    static = True


class ItemsKeyword_2019_09(Keyword, Applicator, ArrayApplicator):
    key = "items"
    instance_types = "array",

    def evaluate(self, instance: JSON, result: Result) -> None:
        if len(instance) == 0:
            return

        elif isinstance(self.json.data, bool):
            self.json.evaluate(instance, result)

        elif isinstance(self.json, JSONSchema):
            for index, item in enumerate(instance):
                self.json.evaluate(item, result)

            if result.passed:
                result.annotate(True)

        elif self.json.type == "array":
            annotation = None
            error = []
            for index, item in enumerate(instance[:len(self.json)]):
                annotation = index
                with result(item, str(index)) as subresult:
                    if not self.json[index].evaluate(item, subresult).passed:
                        error += [index]

            if error:
                result.fail(error)
            else:
                result.annotate(annotation)


class AdditionalItemsKeyword_2019_09(Keyword, Applicator):
    key = "additionalItems"
    instance_types = "array",
    depends_on = "items",

    def evaluate(self, instance: JSON, result: Result) -> None:
        if (items := result.sibling(instance, "items")) and type(items.annotation) is int:
            annotation = None
            error = []
            for index, item in enumerate(instance[(start := items.annotation + 1):], start):
                if self.json.evaluate(item, result).passed:
                    annotation = True
                else:
                    error += [index]
                    # reset to passed for the next iteration
                    result.pass_()

            if error:
                result.fail(error)
            else:
                result.annotate(annotation)

        else:
            result.discard()


class UnevaluatedItemsKeyword_2019_09(Keyword, Applicator):
    key = "unevaluatedItems"
    instance_types = "array",
    depends_on = "items", "additionalItems", "if", "then", "else", "allOf", "anyOf", "oneOf", "not",

    def evaluate(self, instance: JSON, result: Result) -> None:
        last_evaluated_item = -1
        for items_annotation in result.parent.collect_annotations(instance, "items"):
            if items_annotation is True:
                result.discard()
                return
            if type(items_annotation) is int and items_annotation > last_evaluated_item:
                last_evaluated_item = items_annotation

        for additional_items_annotation in result.parent.collect_annotations(instance, "additionalItems"):
            if additional_items_annotation is True:
                result.discard()
                return

        for unevaluated_items_annotation in result.parent.collect_annotations(instance, "unevaluatedItems"):
            if unevaluated_items_annotation is True:
                result.discard()
                return

        annotation = None
        error = []
        for index, item in enumerate(instance[(start := last_evaluated_item + 1):], start):
            if self.json.evaluate(item, result).passed:
                annotation = True
            else:
                error += [index]
                # reset to passed for the next iteration
                result.pass_()

        if error:
            result.fail(error)
        else:
            result.annotate(annotation)



from __future__ import annotations

import collections
import re
import urllib.parse
from typing import Any, Iterable, Mapping, Sequence, TYPE_CHECKING, Union, overload

from jschon.exceptions import JSONPointerError, RelativeJSONPointerError

if TYPE_CHECKING:
    from jschon.json import JSON

__all__ = [
    'JSONPointer',
    'RelativeJSONPointer',
]

JSON_POINTER_RE = '(/([^~/]|(~[01]))*)*'
JSON_INDEX_RE = '0|([1-9][0-9]*)'
RELATIVE_JSON_POINTER_RE = f'(?P<up>{JSON_INDEX_RE})(?P<ref>#|{JSON_POINTER_RE})'


class JSONPointer(Sequence[str]):
    """:rfc:`6901`-conformant JSON Pointer implementation.

    A JSON pointer is a string representing a reference to some value within a
    JSON document. It consists of a series of *reference tokens* each prefixed by
    ``"/"``, each token in turn being the (escaped) JSON object key or the JSON
    array index at the next node down the path to the referenced value. The empty
    string ``""`` represents a reference to an entire JSON document.

    A :class:`JSONPointer` instance is an immutable sequence of the unescaped
    JSON object keys [#]_ and/or array indices [#]_ that comprise the path to a
    referenced value within a JSON document.

    A :class:`JSONPointer` instance is constructed by the concatenation of any
    number of arguments, each of which can be one of:

    - a string conforming to the :rfc:`6901` syntax
    - an iterable of unescaped keys (which may itself be a :class:`JSONPointer`
      instance)

    Two :class:`JSONPointer` instances compare equal if their key sequences are
    identical.

    The ``/`` operator provides a convenient syntax for extending a JSON pointer.
    It produces a new :class:`JSONPointer` instance by copying the left-hand
    argument (a :class:`JSONPointer` instance) and appending the right-hand
    argument (an unescaped key, or an iterable of unescaped keys).

    Taking an index into a :class:`JSONPointer` returns the unescaped key at
    that position. Taking a slice into a :class:`JSONPointer` returns a new
    :class:`JSONPointer` composed of the specified slice of the original's keys.

    .. [#] An *unescaped* object key is the unmodified form of the key within its
       mapping. Keys appearing in an RFC 6901 JSON pointer string, on the other
       hand, are required by the syntax to have reserved characters *escaped*.

    .. [#] Object keys and array indices are uniformly represented as strings
       and referred to as *keys* in the JSONPointer class.
    """

    _json_pointer_re = re.compile(JSON_POINTER_RE)
    _array_index_re = re.compile(JSON_INDEX_RE)

    def __new__(cls, *values: Union[str, Iterable[str]]) -> JSONPointer:
        """Create and return a new :class:`JSONPointer` instance, constructed by
        the concatenation of the given `values`.

        :param values: each value may either be an RFC 6901 string, or an iterable
            of unescaped keys
        :raise JSONPointerError: if a string argument does not conform to the RFC
            6901 syntax
        """
        self = object.__new__(cls)
        self._keys = []

        for value in values:
            if isinstance(value, str):
                if not JSONPointer._json_pointer_re.fullmatch(value):
                    raise JSONPointerError(f"'{value}' is not a valid JSON pointer")
                self._keys.extend(self.unescape(token) for token in value.split('/')[1:])

            elif isinstance(value, JSONPointer):
                self._keys.extend(value._keys)

            elif isinstance(value, Iterable):
                self._keys.extend(value)

            else:
                raise TypeError("Expecting str or Iterable[str]")

        return self

    @overload
    def __getitem__(self, index: int) -> str:
        ...

    @overload
    def __getitem__(self, index: slice) -> JSONPointer:
        ...

    def __getitem__(self, index):
        """Return `self[index]`."""
        if isinstance(index, int):
            return self._keys[index]
        if isinstance(index, slice):
            return JSONPointer(self._keys[index])
        raise TypeError("Expecting int or slice")

    def __len__(self) -> int:
        """Return `len(self)`."""
        return len(self._keys)

    @overload
    def __truediv__(self, suffix: str) -> JSONPointer:
        ...

    @overload
    def __truediv__(self, suffix: Iterable[str]) -> JSONPointer:
        ...

    def __truediv__(self, suffix) -> JSONPointer:
        """Return `self / suffix`."""
        if isinstance(suffix, str):
            return JSONPointer(self, (suffix,))
        if isinstance(suffix, Iterable):
            return JSONPointer(self, suffix)
        return NotImplemented

    def __eq__(self, other: JSONPointer) -> bool:
        """Return `self == other`."""
        if isinstance(other, JSONPointer):
            return self._keys == other._keys
        return NotImplemented

    def __le__(self, other: JSONPointer) -> bool:
        """Return `self <= other`.

        Test whether self is a prefix of other, that is,
        `self == other[:len(self)]`.
        """
        if isinstance(other, JSONPointer):
            return self._keys == other._keys[:len(self._keys)]
        return NotImplemented

    def __lt__(self, other: JSONPointer) -> bool:
        """Return `self < other`.

        Test whether self is a proper prefix of other, that is,
        `self <= other and self != other`.
        """
        if isinstance(other, JSONPointer):
            return len(self._keys) < len(other._keys) and self._keys == other._keys[:len(self._keys)]
        return NotImplemented

    def __hash__(self) -> int:
        """Return `hash(self)`."""
        return hash(tuple(self._keys))

    def __str__(self) -> str:
        """Return `str(self)`."""
        return ''.join([f'/{self.escape(key)}' for key in self._keys])

    def __repr__(self) -> str:
        """Return `repr(self)`."""
        return f"JSONPointer({str(self)!r})"

    def evaluate(self, document: Any) -> Any:
        """Return the value within `document` at the location referenced by `self`.

        `document` may be of any type, though if neither a :class:`Mapping` nor
        a :class:`Sequence`, evaluation by any non-empty :class:`JSONPointer`
        will always fail.

        :param document: any Python object
        :raise JSONPointerError: if `self` references a non-existent location
            within `document`
        """

        def resolve(value, keys):
            from jschon.json import JSON
            if not keys:
                return value

            key = keys.popleft()
            try:
                if (isjson := isinstance(value, JSON)) and value.type == "object" or \
                        not isjson and isinstance(value, Mapping):
                    return resolve(value[key], keys)

                if isjson and value.type == "array" or \
                        not isjson and isinstance(value, Sequence) and \
                        not isinstance(value, str) and \
                        JSONPointer._array_index_re.fullmatch(key):
                    return resolve(value[int(key)], keys)

            except (KeyError, IndexError):
                pass

            raise JSONPointerError(f"Failed to resolve '{self}' against the given document")

        return resolve(document, collections.deque(self._keys))

    @classmethod
    def parse_uri_fragment(cls, value: str) -> JSONPointer:
        """Return a new :class:`JSONPointer` constructed from the :rfc:`6901`
        string obtained by decoding `value`.

        `value` must exclude the initial ``'#'`` of the fragment; this allows
        for sensible interoperation with :class:`~jschon.uri.URI` objects.

        :param value: a percent-encoded URI fragment
        """
        return JSONPointer(urllib.parse.unquote(value))

    def uri_fragment(self) -> str:
        """Return a percent-encoded URI fragment representation of `self`.

        The returned string excludes the initial ``'#'`` of the fragment; this
        allows for sensible interoperation with :class:`~jschon.uri.URI` objects.
        """

        # By default, urllib.parse.quote percent-encodes all characters
        # (apart from '/') defined by the "reserved" production, described
        # in RFC 3986. However, since the "fragment" production allows
        # characters in the "sub-delims" set, we include them in the safe
        # arg.
        #
        # pchar         = unreserved / pct-encoded / sub-delims / ":" / "@"
        # fragment      = *( pchar / "/" / "?" )
        # reserved      = gen-delims / sub-delims
        # gen-delims    = ":" / "/" / "?" / "#" / "[" / "]" / "@"
        # sub-delims    = "!" / "$" / "&" / "'" / "(" / ")"
        #               / "*" / "+" / "," / ";" / "="

        return urllib.parse.quote(str(self), safe="/!$&'()*+,;=")

    @staticmethod
    def escape(key: str) -> str:
        """Return the escaped form of a JSON object key / array index,
        suitable for use in an :rfc:`6901` JSON pointer string.

        :param key: an unescaped key
        """
        return key.replace('~', '~0').replace('/', '~1')

    @staticmethod
    def unescape(token: str) -> str:
        """Return the unescaped form of a reference token appearing in an
        :rfc:`6901` JSON pointer string

        :param token: an RFC 6901 reference token
        """
        return token.replace('~1', '/').replace('~0', '~')


class RelativeJSONPointer:
    _regex = re.compile(RELATIVE_JSON_POINTER_RE)

    def __new__(
            cls,
            value: str = None,
            /,
            *,
            up: int = 0,
            ref: Union[str, JSONPointer] = '',
    ) -> RelativeJSONPointer:
        """Create and return a new :class:`RelativeJSONPointer` instance.

        :param value: a relative JSON pointer-conformant string; if `value` is
            given, keyword args are ignored
        :param up: the number of levels up from the current referenced JSON node
            from which to evaluate `ref`
        :param ref: either the literal ``#``, or a :class:`JSONPointer` instance,
            or a JSON pointer-conformant string
        :raise RelativeJSONPointerError: for any invalid arguments
        """
        self = object.__new__(cls)

        if value is not None:
            if not (match := RelativeJSONPointer._regex.fullmatch(value)):
                raise RelativeJSONPointerError(f"'{value}' is not a valid relative JSON pointer")

            up, ref = match.group('up', 'ref')

        self.up = int(up)
        self.index = ref == '#'

        if self.index:
            self.path = None
        else:
            self.path = JSONPointer(ref) if isinstance(ref, str) else ref

        return self

    def __eq__(self, other: RelativeJSONPointer) -> bool:
        """Return `self == other`."""
        if isinstance(other, RelativeJSONPointer):
            return (self.up == other.up and
                    self.index == other.index and
                    self.path == other.path)
        return NotImplemented

    def __hash__(self) -> int:
        """Return `hash(self)`."""
        return hash((self.up, self.index, self.path))

    def __str__(self) -> str:
        """Return `str(self)`."""
        if self.index:
            return f'{self.up}#'
        return f'{self.up}{self.path}'

    def __repr__(self) -> str:
        """Return `repr(self)`."""
        return f'RelativeJSONPointer({str(self)!r})'

    def evaluate(self, document: JSON) -> Union[int, str, JSON]:
        node = document
        for _ in range(self.up):
            if node.parent is None:
                raise RelativeJSONPointerError('Up too many levels')
            node = node.parent

        if self.index:
            if node.parent is None:
                raise RelativeJSONPointerError('No containing node')
            return int(node.key) if node.parent.type == "array" else node.key

        try:
            return self.path.evaluate(node)
        except JSONPointerError as e:
            raise RelativeJSONPointerError from e



import pathlib
import re

from setuptools import setup, find_packages

HERE = pathlib.Path(__file__).parent.resolve()
README = (HERE / 'README.rst').read_text(encoding='utf-8')
VERSION = eval(re.search(
    '^__version__ = (.*)$',
    (HERE / 'jschon' / '__init__.py').read_text(encoding='utf-8'),
    re.MULTILINE,
)[1])

setup(
    name='jschon',
    version=VERSION,
    description='A pythonic, extensible JSON Schema implementation.',
    long_description=README,
    long_description_content_type='text/x-rst',
    url='https://github.com/marksparkza/jschon',
    author='Mark Jacobson',
    author_email='mark@saeon.ac.za',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    packages=find_packages(exclude=['tests']),
    include_package_data=True,
    python_requires='~=3.8',
    install_requires=[
        'rfc3986',
    ],
    extras_require={
        'requests': [
            'requests',
        ],
        'test': [
            'tox',
        ],
        'dev': [
            'pytest',
            'coverage',
            'hypothesis',
            'pytest-benchmark',
            'pytest-httpserver',
            'requests',
        ],
        'doc': [
            'sphinx',
            'sphinx-rtd-theme',
        ],
    },
)



from __future__ import annotations

import json
from collections import deque
from functools import cached_property
from os import PathLike
from typing import Any, Dict, Iterator, List, Mapping, MutableMapping, MutableSequence, Optional, Sequence, Type, Union

from jschon.jsonpointer import JSONPointer
from jschon.utils import json_dumpf, json_dumps, json_loadf, json_loadr, json_loads

__all__ = [
    'JSON',
    'JSONCompatible',
]

JSONCompatible = Union[None, bool, int, float, str, Sequence[Any], Mapping[str, Any]]
"""Type hint for a JSON-compatible Python object."""


class JSON(MutableSequence['JSON'], MutableMapping[str, 'JSON']):
    """An implementation of the JSON data model."""

    @classmethod
    def loadf(cls, path: Union[str, PathLike], **kwargs: Any) -> JSON:
        """Deserialize a JSON file to a :class:`JSON` instance.

        :param path: the path to the file
        :param kwargs: keyword arguments to pass to the :class:`JSON` (subclass) constructor
        """
        return cls(json_loadf(path), **kwargs)

    @classmethod
    def loadr(cls, url: str, **kwargs: Any) -> JSON:
        """Deserialize a remote JSON resource to a :class:`JSON` instance.

        :param url: the URL of the resource
        :param kwargs: keyword arguments to pass to the :class:`JSON` (subclass) constructor
        """
        return cls(json_loadr(url), **kwargs)

    @classmethod
    def loads(cls, value: str, **kwargs: Any) -> JSON:
        """Deserialize a JSON string to a :class:`JSON` instance.

        :param value: the JSON string
        :param kwargs: keyword arguments to pass to the :class:`JSON` (subclass) constructor
        """
        return cls(json_loads(value), **kwargs)

    def __init__(
            self,
            value: JSONCompatible,
            *,
            parent: JSON = None,
            key: str = None,
            itemclass: Type[JSON] = None,
            **itemkwargs: Any,
    ):
        """Initialize a :class:`JSON` instance from the given JSON-compatible
        `value`.

        The `parent`, `key`, `itemclass` and `itemkwargs` parameters should
        typically only be used in the construction of compound :class:`JSON`
        documents by :class:`JSON` subclasses.

        :param value: a JSON-compatible Python object
        :param parent: the parent node of the instance
        :param key: the index of the instance within its parent
        :param itemclass: the :class:`JSON` subclass used to instantiate
            child nodes of arrays and objects (default: :class:`JSON`)
        :param itemkwargs: keyword arguments to pass to the `itemclass`
            constructor
        """

        self.type: str
        """The JSON type of the instance. One of
        ``null``, ``boolean``, ``number``, ``string``, ``array``, ``object``."""

        self.data: Union[None, bool, int, float, str, List[JSON], Dict[str, JSON]]
        """The instance data.
        
        =========   ===============
        JSON type   data type
        =========   ===============
        null        None
        boolean     bool
        number      int | float
        string      str
        array       list[JSON]
        object      dict[str, JSON]
        =========   ===============
        """

        self.parent: Optional[JSON] = parent
        """The containing JSON instance."""

        self.key: Optional[str] = key
        """The index of the instance within its parent."""

        self.itemclass: Type[JSON] = itemclass or JSON
        """The :class:`JSON` class type of child instances."""

        self.itemkwargs: Dict[str, Any] = itemkwargs
        """Keyword arguments to the :attr:`itemclass` constructor."""

        if value is None:
            self.type = "null"
            self.data = None

        elif isinstance(value, bool):
            self.type = "boolean"
            self.data = value

        elif isinstance(value, (int, float)):
            self.type = "number"
            self.data = value

        elif isinstance(value, str):
            self.type = "string"
            self.data = value

        elif isinstance(value, Sequence):
            self.type = "array"
            self.data = [
                self.itemclass(v, parent=self, key=str(i), **self.itemkwargs)
                for i, v in enumerate(value)
            ]

        elif isinstance(value, Mapping):
            self.type = "object"
            self.data = {
                k: self.itemclass(v, parent=self, key=k, **self.itemkwargs)
                for k, v in value.items()
            }

        else:
            raise TypeError(f"{value=} is not JSON-compatible")

    @cached_property
    def path(self) -> JSONPointer:
        """Return the path to the instance from the document root."""
        keys = deque()
        node = self
        while node.parent is not None:
            keys.appendleft(node.key)
            node = node.parent
        return JSONPointer(keys)

    @cached_property
    def value(self) -> JSONCompatible:
        """Return the instance data as a JSON-compatible Python object."""
        if isinstance(self.data, list):
            return [item.value for item in self.data]
        if isinstance(self.data, dict):
            return {key: item.value for key, item in self.data.items()}
        return self.data

    def _invalidate_path(self) -> None:
        try:
            del self.path
        except AttributeError:
            pass
        if self.type == 'array':
            for item in self.data:
                item._invalidate_path()
        elif self.type == 'object':
            for item in self.data.values():
                item._invalidate_path()

    def _invalidate_value(self) -> None:
        try:
            del self.value
        except AttributeError:
            pass
        if self.parent is not None:
            self.parent._invalidate_value()

    def dumpf(self, path: Union[str, PathLike]) -> None:
        """Serialize the instance data to a JSON file.

        :param path: the path to the file
        """
        json_dumpf(self.data, path)

    def dumps(self) -> str:
        """Serialize the instance data to a JSON string."""
        return json_dumps(self.data)

    def __repr__(self) -> str:
        """Return `repr(self)`."""
        return f'{self.__class__.__name__}({json.loads(self.dumps())!r})'

    def __str__(self) -> str:
        """Return `str(self)`."""
        return self.dumps()

    def __bool__(self) -> bool:
        """Return `bool(self)`."""
        return bool(self.data)

    def __len__(self) -> int:
        """Return `len(self)`.

        Supported for JSON types ``string``, ``array`` and ``object``.
        """
        return len(self.data)

    def __iter__(self) -> Iterator:
        """Return `iter(self)`.

        Supported for JSON types ``array`` and ``object``.
        """
        return iter(self.data)

    def __getitem__(self, index: Union[int, slice, str]) -> JSON:
        """Return `self[index]`.

        Supported for JSON types ``array`` and ``object``.
        """
        return self.data[index]

    def __setitem__(self, index: Union[int, str], obj: Union[JSON, JSONCompatible]) -> None:
        """Set `self[index]` to `obj`.

        Supported for JSON types ``array`` and ``object``.
        """
        self.data[index] = self.itemclass(
            obj.value if isinstance(obj, JSON) else obj,
            parent=self,
            key=str(index),
            **self.itemkwargs,
        )
        self._invalidate_value()

    def __delitem__(self, index: Union[int, str]) -> None:
        """Delete `self[index]`.

        Supported for JSON types ``array`` and ``object``.
        """
        del self.data[index]
        self._invalidate_value()
        if self.type == 'array':
            for item in self.data[index:]:
                item.key = str(int(item.key) - 1)
                item._invalidate_path()

    def insert(self, index: int, obj: Union[JSON, JSONCompatible]) -> None:
        """Insert `obj` before `index`.

        Supported for JSON type ``array``.
        """
        self.data.insert(index, self.itemclass(
            obj.value if isinstance(obj, JSON) else obj,
            parent=self,
            key=str(index),
            **self.itemkwargs,
        ))
        self._invalidate_value()
        for item in self.data[index + 1:]:
            item.key = str(int(item.key) + 1)
            item._invalidate_path()

    def __eq__(self, other: Union[JSON, JSONCompatible]) -> bool:
        """Return `self == other`."""
        if not isinstance(other, JSON):
            other = JSON(other)
        if self.type == other.type:
            if self.type == "array":
                return len(self) == len(other) and \
                       all(item == other[i] for i, item in enumerate(self))
            if self.type == "object":
                return self.keys() == other.keys() and \
                       all(item == other[k] for k, item in self.items())
            return self.data == other.data
        return NotImplemented

    def __ge__(self, other: Union[JSON, int, float, str]) -> bool:
        """Return `self >= other`.

        Supported for JSON types ``number`` and ``string``.
        """
        if isinstance(other, JSON):
            return self.data >= other.data
        return self.data >= other

    def __gt__(self, other: Union[JSON, int, float, str]) -> bool:
        """Return `self > other`.

        Supported for JSON types ``number`` and ``string``.
        """
        if isinstance(other, JSON):
            return self.data > other.data
        return self.data > other

    def __le__(self, other: Union[JSON, int, float, str]) -> bool:
        """Return `self <= other`.

        Supported for JSON types ``number`` and ``string``.
        """
        if isinstance(other, JSON):
            return self.data <= other.data
        return self.data <= other

    def __lt__(self, other: Union[JSON, int, float, str]) -> bool:
        """Return `self < other`.

        Supported for JSON types ``number`` and ``string``.
        """
        if isinstance(other, JSON):
            return self.data < other.data
        return self.data < other



from __future__ import annotations

import inspect
from typing import Any, Dict, Mapping, Optional, Sequence, TYPE_CHECKING, Tuple, Type

from jschon.json import JSON, JSONCompatible
from jschon.jsonschema import JSONSchema, Result
from jschon.uri import URI

if TYPE_CHECKING:
    from jschon.catalog import Catalog

__all__ = [
    'Metaschema',
    'Vocabulary',
    'Keyword',
    'KeywordClass',
    'ApplicatorMixin',
    'Applicator',
    'ArrayApplicator',
    'PropertyApplicator',
]


class Metaschema(JSONSchema):
    """A metaschema declares the set of vocabularies that are available
    to any schema which references it, and provides any such schema with
    its :class:`Keyword` classes.
    
    :class:`Metaschema` is itself a subclass of :class:`~jschon.jsonschema.JSONSchema`,
    and may be used to validate any referencing schema.
    """

    def __init__(
            self,
            catalog: Catalog,
            value: Mapping[str, JSONCompatible],
            core_vocabulary: Vocabulary,
            *default_vocabularies: Vocabulary,
            **kwargs: Any,
    ):
        self.core_vocabulary: Vocabulary = core_vocabulary
        self.default_vocabularies: Tuple[Vocabulary, ...] = default_vocabularies
        self.kwclasses: Dict[str, KeywordClass] = {}
        super().__init__(value, catalog=catalog, session='__meta__', **kwargs)

    def _bootstrap(self, value: Mapping[str, JSONCompatible]) -> None:
        super()._bootstrap(value)
        self.kwclasses.update(self.core_vocabulary.kwclasses)
        if "$vocabulary" not in value:
            for vocabulary in self.default_vocabularies:
                self.kwclasses.update(vocabulary.kwclasses)


class Vocabulary:
    """A vocabulary declares a set of keywords that may be used in the
    evaluation of JSON documents, and provides a runtime implementation
    (in the form of a :class:`Keyword` class) for each of those keywords."""

    def __init__(self, uri: URI, *kwclasses: KeywordClass):
        self.uri: URI = uri
        self.kwclasses: Dict[str, KeywordClass] = {
            kwclass.key: kwclass for kwclass in kwclasses
        }


class Keyword:
    key: str = ...

    instance_types: Tuple[str, ...] = "null", "boolean", "integer", "number", "string", "array", "object",
    """The types of instance that the keyword can evaluate."""

    depends_on: Tuple[str, ...] = ()
    """Keywords that must be evaluated before this keyword can be evaluated."""

    static: bool = False

    def __init__(self, parentschema: JSONSchema, value: JSONCompatible):
        for base_cls in inspect.getmro(self.__class__):
            if issubclass(base_cls, ApplicatorMixin):
                if (kwjson := base_cls.jsonify(parentschema, self.key, value)) is not None:
                    break
        else:
            kwjson = JSON(value, parent=parentschema, key=self.key)

        self.json: JSON = kwjson
        self.parentschema: JSONSchema = parentschema

    def can_evaluate(self, instance: JSON) -> bool:
        if instance.type in self.instance_types:
            return True

        if instance.type == "number" and "integer" in self.instance_types:
            return instance.data == int(instance.data)

        return False

    def evaluate(self, instance: JSON, result: Result) -> None:
        pass

    def __str__(self) -> str:
        return f'{self.json.path}: {self.json}'


KeywordClass = Type[Keyword]


class ApplicatorMixin:
    @classmethod
    def jsonify(cls, parentschema: JSONSchema, key: str, value: JSONCompatible) -> Optional[JSON]:
        raise NotImplementedError


class Applicator(ApplicatorMixin):
    """A :class:`~Keyword` class mixin that sets up a subschema for
    an applicator keyword."""

    @classmethod
    def jsonify(cls, parentschema: JSONSchema, key: str, value: JSONCompatible) -> Optional[JSON]:
        if isinstance(value, (bool, Mapping)):
            return JSONSchema(
                value,
                parent=parentschema,
                key=key,
                catalog=parentschema.catalog,
                session=parentschema.session,
            )


class ArrayApplicator(ApplicatorMixin):
    """A :class:`~Keyword` class mixin that sets up an array of subschemas
    for an applicator keyword."""

    @classmethod
    def jsonify(cls, parentschema: JSONSchema, key: str, value: JSONCompatible) -> Optional[JSON]:
        if isinstance(value, Sequence):
            return JSON(
                value,
                parent=parentschema,
                key=key,
                itemclass=JSONSchema,
                catalog=parentschema.catalog,
                session=parentschema.session,
            )


class PropertyApplicator(ApplicatorMixin):
    """A :class:`~Keyword` class mixin that sets up property-based subschemas
    for an applicator keyword."""

    @classmethod
    def jsonify(cls, parentschema: JSONSchema, key: str, value: JSONCompatible) -> Optional[JSON]:
        if isinstance(value, Mapping):
            return JSON(
                value,
                parent=parentschema,
                key=key,
                itemclass=JSONSchema,
                catalog=parentschema.catalog,
                session=parentschema.session,
            )



from typing import Any, Callable, Dict

from jschon.json import JSONCompatible
from jschon.jsonschema import Result

__all__ = [
    'OutputFormatter',
    'output_formatter',
    'create_output',
]

OutputFormatter = Callable[..., JSONCompatible]
"""Call signature for a function decorated with :func:`output_formatter`.

The function must take a :class:`~jschon.jsonschema.Result` object
as its first argument.
"""

_formatters: Dict[str, OutputFormatter] = {}


def output_formatter(name: str = None):
    """A decorator for a JSON Schema output formatting function.

    :param name: The format name; defaults to the name of the decorated
        function.
    """

    def decorator(f):
        formatter_name = name if isinstance(name, str) else f.__name__
        _formatters[formatter_name] = f
        return f

    return decorator(name) if callable(name) else decorator


def create_output(result: Result, format: str, **kwargs: Any) -> JSONCompatible:
    return _formatters[format](result, **kwargs)


@output_formatter
def flag(result: Result) -> JSONCompatible:
    return {
        "valid": result.valid
    }


@output_formatter
def basic(result: Result) -> JSONCompatible:
    def visit(node: Result):
        if node.valid is valid:
            if (msgval := getattr(node, msgkey)) is not None:
                yield {
                    "instanceLocation": str(node.instance.path),
                    "keywordLocation": str(node.path),
                    "absoluteKeywordLocation": str(node.absolute_uri),
                    msgkey: msgval,
                }
            for child in node.children.values():
                yield from visit(child)

    valid = result.valid
    msgkey = "annotation" if valid else "error"
    childkey = "annotations" if valid else "errors"

    return {
        "valid": valid,
        childkey: [output for output in visit(result)],
    }


@output_formatter
def detailed(result: Result) -> JSONCompatible:
    def visit(node: Result):
        output = {
            "instanceLocation": str(node.instance.path),
            "keywordLocation": str(node.path),
            "absoluteKeywordLocation": str(node.absolute_uri),
            childkey: [visit(child) for child in node.children.values()
                       if child.valid is valid],
        }
        if not output[childkey]:
            del output[childkey]
            if (msgval := getattr(node, msgkey)) is not None:
                output[msgkey] = msgval
        elif len(output[childkey]) == 1:
            output = output[childkey][0]

        return output

    valid = result.valid
    msgkey = "annotation" if valid else "error"
    childkey = "annotations" if valid else "errors"

    return {
        "valid": valid,
        "instanceLocation": str(result.instance.path),
        "keywordLocation": str(result.path),
        "absoluteKeywordLocation": str(result.absolute_uri),
        childkey: [visit(child) for child in result.children.values()
                   if child.valid is valid],
    }


@output_formatter
def verbose(result: Result) -> JSONCompatible:
    def visit(node: Result):
        output = {
            "valid": (valid := node.valid),
            "instanceLocation": str(node.instance.path),
            "keywordLocation": str(node.path),
            "absoluteKeywordLocation": str(node.absolute_uri),
        }

        msgkey = "annotation" if valid else "error"
        if (msgval := getattr(node, msgkey)) is not None:
            output[msgkey] = msgval

        childkey = "annotations" if valid else "errors"
        if childarr := [visit(child) for child in node.children.values()]:
            output[childkey] = childarr

        return output

    return visit(result)
